from __future__ import division
from functools import partial
import numpy as np
from menpo.shape import PointCloud
from menpo.transform import (
    scale_about_centre, rotate_ccw_about_centre, Translation,
    Scale, AlignmentAffine, AlignmentSimilarity)
import menpofit.checks as checks
from menpofit.visualize import print_progress


def noisy_alignment_similarity_transform(source, target, noise_type='uniform',
                                         noise_percentage=0.1, rotation=False):
    r"""
    Constructs and perturbs the optimal similarity transform between source
    and target by adding noise to its parameters.

    Parameters
    ----------
    source: :class:`menpo.shape.PointCloud`
        The source pointcloud instance used in the alignment
    target: :class:`menpo.shape.PointCloud`
        The target pointcloud instance used in the alignment
    noise_type: str, optional
        The type of noise to be added, 'uniform' or 'gaussian'.
    noise_percentage: 0 < float < 1 or triplet of 0 < float < 1, optional
        The standard percentage of noise to be added. If float the same amount
        of noise is applied to the scale, rotation and translation
        parameters of the true similarity transform. If triplet of
        floats, the first, second and third elements denote the amount of
        noise to be applied to the scale, rotation and translation
        parameters respectively.
    rotation: boolean, optional
        If False rotation is not considered when computing the optimal
        similarity transform between source and target.

    Returns
    -------
    noisy_alignment_similarity_transform : :class: `menpo.transform.Similarity`
        The noisy Similarity Transform between source and target.
    """
    if isinstance(noise_percentage, float):
        noise_percentage = [noise_percentage] * 3
    elif len(noise_percentage) == 1:
        noise_percentage *= 3

    similarity = AlignmentSimilarity(source, target, rotation=rotation)

    if noise_type is 'gaussian':
        s = noise_percentage[0] * (0.5 / 3) * np.asscalar(np.random.randn(1))
        r = noise_percentage[1] * (180 / 3) * np.asscalar(np.random.randn(1))
        t = noise_percentage[2] * (target.range() / 3) * np.random.randn(2)

        s = scale_about_centre(target, 1 + s)
        r = rotate_ccw_about_centre(target, r)
        t = Translation(t, source.n_dims)
    elif noise_type is 'uniform':
        s = noise_percentage[0] * 0.5 * (2 * np.asscalar(np.random.randn(1)) - 1)
        r = noise_percentage[1] * 180 * (2 * np.asscalar(np.random.rand(1)) - 1)
        t = noise_percentage[2] * target.range() * (2 * np.random.rand(2) - 1)

        s = scale_about_centre(target, 1. + s)
        r = rotate_ccw_about_centre(target, r)
        t = Translation(t, source.n_dims)
    else:
        raise ValueError('Unexpected noise type. '
                         'Supported values are {gaussian, uniform}')

    return similarity.compose_after(t.compose_after(s.compose_after(r)))


def noisy_target_alignment_transform(source, target,
                                     alignment_transform_cls=AlignmentAffine,
                                     noise_std=0.1, **kwargs):
    r"""
    Constructs and the optimal alignment transform between the source and
    a noisy version of the target obtained by adding white noise to each of
    its points.

    Parameters
    ----------
    source: :class:`menpo.shape.PointCloud`
        The source pointcloud instance used in the alignment
    target: :class:`menpo.shape.PointCloud`
        The target pointcloud instance used in the alignment
    alignment_transform_cls: :class:`menpo.transform.Alignment`, optional
        The alignment transform class used to perform the alignment.
    noise_std: float or triplet of floats, optional
        The standard deviation of the white noise to be added to each one of
        the target points.

    Returns
    -------
    noisy_transform : :class: `menpo.transform.Alignment`
        The noisy Similarity Transform
    """
    noise = noise_std * target.range() * np.random.randn(target.n_points,
                                                         target.n_dims)
    noisy_target = PointCloud(target.points + noise)
    return alignment_transform_cls(source, noisy_target, **kwargs)


def noisy_shape_from_bounding_box(shape, bounding_box, noise_type='uniform',
                                  noise_percentage=0.05, rotation=False):
    transform = noisy_alignment_similarity_transform(
        shape.bounding_box(), bounding_box, noise_type=noise_type,
        noise_percentage=noise_percentage, rotation=rotation)
    return transform.apply(shape)


def noisy_shape_from_shape(reference_shape, shape, noise_type='uniform',
                           noise_percentage=0.05, rotation=False):
    transform = noisy_alignment_similarity_transform(
        reference_shape, shape, noise_type=noise_type,
        noise_percentage=noise_percentage, rotation=rotation)
    return transform.apply(reference_shape)


def align_shape_with_bounding_box(shape, bounding_box,
                                  alignment_transform_cls=AlignmentSimilarity,
                                  **kwargs):
    r"""
    Aligns the shape with the bounding box using a particular ali .

    Parameters
    ----------
    source: :class:`menpo.shape.PointCloud`
        The shape instance used in the alignment.
    bounding_box: :class:`menpo.shape.PointCloud`
        The bounding box instance used in the alignment.
    alignment_transform_cls: :class:`menpo.transform.Alignment`, optional
        The class of the alignment transform used to perform the alignment.

    Returns
    -------
    noisy_transform : :class: `menpo.transform.Alignment`
        The noisy Alignment Transform
    """
    shape_bb = shape.bounding_box()
    transform = alignment_transform_cls(shape_bb, bounding_box, **kwargs)
    return transform.apply(shape)


class MultiFitter(object):
    r"""
    """
    @property
    def n_scales(self):
        r"""
        The number of scales used during alignment.

        :type: `int`
        """
        return len(self.scales)

    def fit_from_shape(self, image, initial_shape, max_iters=20, gt_shape=None,
                       crop_image=None, **kwargs):
        r"""
        Fits the multilevel fitter to an image.

        Parameters
        -----------
        image: :map:`Image` or subclass
            The image to be fitted.
        initial_shape: :map:`PointCloud`
            The initial shape estimate from which the fitting procedure
            will start.
        max_iters: `int` or `list` of `int`, optional
            The maximum number of iterations.
            If `int`, specifies the overall maximum number of iterations.
            If `list` of `int`, specifies the maximum number of iterations per
            level.
        gt_shape: :map:`PointCloud`
            The ground truth shape associated to the image.
        crop_image: `None` or float`, optional
            If `float`, it specifies the proportion of the border wrt the
            initial shape to which the image will be internally cropped around
            the initial shape range.
            If `None`, no cropping is performed.

            This will limit the fitting algorithm search region but is
            likely to speed up its running time, specially when the
            modeled object occupies a small portion of the image.
        **kwargs:
            Additional keyword arguments that can be passed to specific
            implementations of ``_fit`` method.

        Returns
        -------
        multi_fitting_result: :map:`MultilevelFittingResult`
            The multilevel fitting result containing the result of
            fitting procedure.
        """
        # generate the list of images to be fitted
        images, initial_shapes, gt_shapes = self._prepare_image(
            image, initial_shape, gt_shape=gt_shape, crop_image=crop_image)

        # work out the affine transform between the initial shape of the
        # highest pyramidal level and the initial shape of the original image
        affine_correction = AlignmentAffine(initial_shapes[-1], initial_shape)

        # run multilevel fitting
        algorithm_results = self._fit(images, initial_shapes[0],
                                      max_iters=max_iters,
                                      gt_shapes=gt_shapes, **kwargs)

        # build multilevel fitting result
        fitter_result = self._fitter_result(
            image, algorithm_results, affine_correction, gt_shape=gt_shape)

        return fitter_result

    def fit_from_bb(self, image, bounding_box, max_iters=20, gt_shape=None,
                    crop_image=None, **kwargs):
        initial_shape = align_shape_with_bounding_box(self.reference_shape,
                                                      bounding_box)
        return self.fit_from_shape(image, initial_shape, max_iters=max_iters,
                                   gt_shape=gt_shape, crop_image=crop_image,
                                   **kwargs)

    def _prepare_image(self, image, initial_shape, gt_shape=None,
                       crop_image=0.5):
        r"""
        Prepares the image to be fitted.

        The image is first rescaled wrt the ``reference_landmarks`` and then
        a gaussian pyramid is applied. Depending on the
        ``pyramid_on_features`` flag, the pyramid is either applied to the
        features image computed from the rescaled imaged or applied to the
        rescaled image and features extracted at each pyramidal level.

        Parameters
        ----------
        image : :map:`Image` or subclass
            The image to be fitted.
        initial_shape : :map:`PointCloud`
            The initial shape from which the fitting will start.
        gt_shape : class : :map:`PointCloud`, optional
            The original ground truth shape associated to the image.
        crop_image: `None` or float`, optional
            If `float`, it specifies the proportion of the border wrt the
            initial shape to which the image will be internally cropped around
            the initial shape range.
            If `None`, no cropping is performed.
            This will limit the fitting algorithm search region but is
            likely to speed up its running time, specially when the
            modeled object occupies a small portion of the image.
        Returns
        -------
        images : `list` of :map:`Image` or subclass
            The list of images that will be fitted by the fitters.
        initial_shapes : `list` of :map:`PointCloud`
            The initial shape for each one of the previous images.
        gt_shapes : `list` of :map:`PointCloud`
            The ground truth shape for each one of the previous images.
        """
        # Attach landmarks to the image
        image.landmarks['__initial_shape'] = initial_shape
        if gt_shape:
            image.landmarks['__gt_shape'] = gt_shape

        if crop_image:
            # If specified, crop the image
            image = image.crop_to_landmarks_proportion(crop_image,
                                                       group='__initial_shape')

        # Rescale image wrt the scale factor between reference_shape and
        # initial_shape
        image = image.rescale_to_pointcloud(self.reference_shape,
                                            group='__initial_shape')

        # Compute image representation
        images = []
        for i in range(self.n_scales):
            # Handle features
            if i == 0 or self.holistic_features[i] is not self.holistic_features[i - 1]:
                # Compute features only if this is the first pass through
                # the loop or the features at this scale are different from
                # the features at the previous scale
                feature_image = self.holistic_features[i](image)

            # Handle scales
            if self.scales[i] != 1:
                # Scale feature images only if scale is different than 1
                scaled_image = feature_image.rescale(self.scales[i])
            else:
                scaled_image = feature_image

            # Add scaled image to list
            images.append(scaled_image)

        # Get initial shapes per level
        initial_shapes = [i.landmarks['__initial_shape'].lms for i in images]

        # Get ground truth shapes per level
        if gt_shape:
            gt_shapes = [i.landmarks['__gt_shape'].lms for i in images]
        else:
            gt_shapes = None

        # detach added landmarks from image
        del image.landmarks['__initial_shape']
        if gt_shape:
            del image.landmarks['__gt_shape']

        return images, initial_shapes, gt_shapes

    def _fit(self, images, initial_shape, gt_shapes=None, max_iters=20,
             **kwargs):
        r"""
        Fits the fitter to the multilevel pyramidal images.

        Parameters
        -----------
        images: :class:`menpo.image.masked.MaskedImage` list
            The images to be fitted.
        initial_shape: :class:`menpo.shape.PointCloud`
            The initial shape from which the fitting will start.
        gt_shapes: :class:`menpo.shape.PointCloud` list, optional
            The original ground truth shapes associated to the multilevel
            images.
        max_iters: int or list, optional
            The maximum number of iterations.
            If int, then this will be the overall maximum number of iterations
            for all the pyramidal levels.
            If list, then a maximum number of iterations is specified for each
            pyramidal level.

        Returns
        -------
        algorithm_results: :class:`FittingResult` list
            The fitting object containing the state of the whole fitting
            procedure.
        """
        # Perform check
        max_iters = checks.check_max_iters(max_iters, self.n_scales)

        # Set initial and ground truth shapes
        shape = initial_shape
        gt_shape = None

        # Initialize list of algorithm results
        algorithm_results = []
        for i in range(self.n_scales):
            # Handle ground truth shape
            if gt_shapes is not None:
                gt_shape = gt_shapes[i]

            # Run algorithm
            algorithm_result = self.algorithms[i].run(images[i], shape,
                                                      gt_shape=gt_shape,
                                                      max_iters=max_iters[i],
                                                      **kwargs)
            # Add algorithm result to the list
            algorithm_results.append(algorithm_result)

            # Prepare this scale's final shape for the next scale
            shape = algorithm_result.final_shape
            if self.scales[i] != self.scales[-1]:
                shape = Scale(self.scales[i + 1] / self.scales[i],
                              n_dims=shape.n_dims).apply(shape)

        # Return list of algorithm results
        return algorithm_results


# TODO: document me!
class ModelFitter(MultiFitter):
    r"""
    """
    @property
    def reference_shape(self):
        r"""
        The reference shape of the AAM.

        :type: :map:`PointCloud`
        """
        return self._model.reference_shape

    @property
    def holistic_features(self):
        r"""
        """
        return self._model.holistic_features

    @property
    def scales(self):
        return self._model.scales

    def _check_n_shape(self, n_shape):
        checks.set_models_components(self._model.shape_models, n_shape)

    def perturb_from_bb(self, gt_shape, bb,
                        perturb_func=noisy_shape_from_bounding_box):
        return perturb_func(gt_shape, bb)

    def perturb_from_gt_bb(self, gt_bb,
                           perturb_func=noisy_shape_from_bounding_box):
        return perturb_func(gt_bb, gt_bb)


def generate_perturbations_from_gt(images, n_perturbations, perturb_func,
                                   gt_group=None, bb_group_glob=None,
                                   verbose=False):

    if bb_group_glob is None:
        bb_generator = lambda im: [im.landmarks[gt_group].lms.bounding_box()]
        n_bbs = 1
    else:
        def bb_glob(im):
            for k, v in im.landmarks.items_matching(bb_group_glob):
                yield v.lms.bounding_box()
        bb_generator = bb_glob
        n_bbs = len(list(bb_glob(images[0])))

    if n_bbs == 0:
        raise ValueError('Must provide a valid bounding box glob - no bounding '
                         'boxes matched the following '
                         'glob: {}'.format(bb_group_glob))

    # If we have multiple boxes - we didn't just throw them away, we re-add them
    # to the end
    if bb_group_glob is not None:
        msg = '- Generating {0} ({1} perturbations * {2} provided boxes) new ' \
              'initial bounding boxes + {2} provided boxes per image'.format(
            n_perturbations * n_bbs, n_perturbations, n_bbs)
    else:
        msg = '- Generating {} new bounding boxes directly from the ' \
              'ground truth shape'.format(n_perturbations)

    wrap = partial(print_progress, prefix=msg, verbose=verbose)
    for im in wrap(images):
        gt_s = im.landmarks[gt_group].lms.bounding_box()

        k = 0
        for bb in bb_generator(im):
            for _ in range(n_perturbations):
                p_s = perturb_func(gt_s, bb).bounding_box()
                perturb_bbox_group = '__generated_bb_{}'.format(k)
                im.landmarks[perturb_bbox_group] = p_s
                k += 1

            if bb_group_glob is not None:
                perturb_bbox_group = '__generated_bb_{}'.format(k)
                im.landmarks[perturb_bbox_group] = bb
                k += 1

        if im.has_landmarks_outside_bounds:
            im.constrain_landmarks_to_bounds()

    generated_bb_func = lambda x: [v.lms for k, v in x.landmarks.items_matching(
        '__generated_bb_*')]
    return generated_bb_func



from __future__ import division
import itertools
import numpy as np


def batch(iterable, n):
    it = iter(iterable)
    while True:
        chunk = tuple(itertools.islice(it, n))
        if not chunk:
            return
        yield chunk


def build_grid(shape):
    r"""
    """
    shape = np.asarray(shape)
    half_shape = np.floor(shape / 2)
    half_shape = np.require(half_shape, dtype=int)
    start = -half_shape
    end = half_shape + shape % 2
    sampling_grid = np.mgrid[start[0]:end[0], start[1]:end[1]]
    return np.rollaxis(sampling_grid, 0, 3)



from __future__ import division
from menpofit.result import ParametricAlgorithmResult, MultiFitterResult


# TODO: document me!
class AAMAlgorithmResult(ParametricAlgorithmResult):
    r"""
    """
    def __init__(self, image, algorithm, shape_parameters, cost_functions=None,
                 appearance_parameters=None, gt_shape=None):
        super(AAMAlgorithmResult, self).__init__(
            image, algorithm, shape_parameters, gt_shape=gt_shape)
        self._cost_functions = cost_functions
        self.appearance_parameters = appearance_parameters
        self._warped_images = None
        self._appearance_reconstructions = None
        self._costs = None

    @property
    def warped_images(self):
        if self._warped_images is None:
            self._warped_images = []
            for p in self.shape_parameters:
                self.algorithm.transform._from_vector_inplace(p)
                self._warped_images.append(
                    self.algorithm.warp(self.image))
        return self._warped_images

    @property
    def appearance_reconstructions(self):
        if self.appearance_parameters is not None:
            if self._appearance_reconstructions is None:
                self._appearance_reconstructions = []
                for c in self.appearance_parameters:
                    instance = self.algorithm.appearance_model.instance(c)
                    self._appearance_reconstructions.append(instance)
            return self._appearance_reconstructions
        else:
            raise ValueError('appearance_reconstructions is not well '
                             'defined for the chosen AAM algorithm: '
                             '{}'.format(self.algorithm.__class__))

    @property
    def costs(self):
        if self._cost_functions is not None:
            if self._costs is None:
                self._costs = [f() for f in self._cost_functions]
            return self._costs
        else:
            raise ValueError('costs is not well '
                             'defined for the chosen AAM algorithm: '
                             '{}'.format(self.algorithm.__class__))


# TODO: document me!
class LinearAAMAlgorithmResult(AAMAlgorithmResult):
    r"""
    """
    @property
    def shapes(self, as_points=False):
        return [self.algorithm.transform.from_vector(p).sparse_target
                for p in self.shape_parameters]

    @property
    def final_shape(self):
        return self.final_transform.sparse_target

    @property
    def initial_shape(self):
        return self.initial_transform.sparse_target


# TODO: document me!
class AAMFitterResult(MultiFitterResult):
    r"""
    """
    def __init__(self, image, fitter, algorithm_results, affine_correction,
                 gt_shape=None):
        super(AAMFitterResult, self).__init__(
            image, fitter, algorithm_results, affine_correction,
            gt_shape=gt_shape)
        self._warped_images = None

    @property
    def warped_images(self):
        if self._warped_images is None:
            algorithm = self.algorithm_results[-1].algorithm
            self._warped_images = []
            for s in self.shapes:
                algorithm.transform.set_target(s)
                self._warped_images.append(
                    algorithm.warp(self.image))
        return self._warped_images

    @property
    def appearance_reconstructions(self):
        reconstructions = self.algorithm_results[0].appearance_reconstructions
        if reconstructions is not None:
            for a in self.algorithm_results[1:]:
                reconstructions = (reconstructions +
                                   a.appearance_reconstructions)
        return reconstructions

    @property
    def costs(self):
        costs = []
        for a in self.algorithm_results:
            costs += a.costs
        return costs



from .fitter import LucasKanadeFitter
from .algorithm import (
    ForwardAdditive, ForwardCompositional, InverseCompositional)
from .residual import (
    SSD, FourierSSD, ECC, GradientImages, GradientCorrelation)



import numpy as np

from menpo.math import pca


class IRLRegression(object):
    r"""
    Incremental Regularized Linear Regression
    """
    def __init__(self, alpha=0, bias=True, incrementable=False):
        self.alpha = alpha
        self.bias = bias
        self.incrementable = incrementable
        self.V = None
        self.W = None

    def train(self, X, Y):
        if self.bias:
            # add bias
            X = np.hstack((X, np.ones((X.shape[0], 1))))

        # regularized linear regression
        XX = X.T.dot(X)
        # ensure covariance is perfectly symmetric for inversion
        XX = (XX + XX.T) / 2.0
        if self.alpha:
            np.fill_diagonal(XX, self.alpha + np.diag(XX))
        if self.incrementable:
            self.V = np.linalg.inv(XX)
        self.W = np.linalg.solve(XX, X.T.dot(Y))

    def increment(self, X, Y):
        if not self.incrementable:
            raise ValueError('Model is not incrementable')

        if self.bias:
            # add bias
            X = np.hstack((X, np.ones((X.shape[0], 1))))

        # incremental regularized linear regression
        U = X.dot(self.V).dot(X.T)
        if self.alpha:
            np.fill_diagonal(U, self.alpha + np.diag(U))
        U = np.linalg.inv(U)
        Q = self.V.dot(X.T).dot(U).dot(X)
        self.V = self.V - Q.dot(self.V)
        self.W = self.W - Q.dot(self.W) + self.V.dot(X.T.dot(Y))

    def predict(self, x):
        if self.bias:
            if len(x.shape) == 1:
                x = np.hstack((x, np.ones(1)))
            else:
                x = np.hstack((x, np.ones((x.shape[0], 1))))
        return np.dot(x, self.W)


# TODO: document me!
class IIRLRegression(IRLRegression):
    r"""
    Indirect Incremental Regularized Linear Regression
    """
    def __init__(self, alpha=0, bias=False, alpha2=0):
        # TODO: Can we model the bias? May need to slice off of prediction?
        super(IIRLRegression, self).__init__(alpha=alpha, bias=False)
        self.alpha2 = alpha2

    def train(self, X, Y):
        # regularized linear regression exchanging the roles of X and Y
        super(IIRLRegression, self).train(Y, X)
        J = self.W
        # solve the original problem by computing the pseudo-inverse of the
        # previous solution
        # Note that everything is transposed from the above exchanging of roles
        H = J.dot(J.T)
        if self.alpha2:
            np.fill_diagonal(H, self.alpha2 + np.diag(H))
        self.W = np.linalg.solve(H, J).T

    def increment(self, X, Y):
        # incremental least squares exchanging the roles of X and Y
        super(IIRLRegression, self).increment(Y, X)
        J = self.W
        # solve the original problem by computing the pseudo-inverse of the
        # previous solution
        # Note that everything is transposed from the above exchanging of roles
        H = J.dot(J.T)
        if self.alpha2:
            np.fill_diagonal(H, self.alpha2 + np.diag(H))
        self.W = np.linalg.solve(H, J)

    def predict(self, x):
        return np.dot(x, self.W)


class PCRRegression(object):
    r"""
    Multivariate Linear Regression using Principal Component Regression

    Parameters
    ----------
    X : numpy.array
        The regression features used to create the coefficient matrix.
    T : numpy.array
        The shapes differential that denote the dependent variable.
    variance: float or None, Optional
        The SVD variance.
        Default: None

    Raises
    ------
    ValueError
        variance must be set to a number between 0 and 1
    """
    def __init__(self, variance=None, bias=True, eps=1e-10):
        self.variance = variance
        self.bias = bias
        self.R = None
        self.V = None
        self.eps = eps

    def train(self, X, Y):
        if self.bias:
            X = np.hstack((X, np.ones((X.shape[0], 1))))

        # Reduce variance
        U, s, self.V = np.linalg.svd(X, full_matrices=False)
        if self.variance:
            variation = np.cumsum(s) / np.sum(s)
            # Inverted for easier parameter semantics
            k = np.sum(variation < self.variance)
            U = U[:, :k]
            self.V = self.V[:k, :]
            s = s[:k]
        S = np.diag(s)

        # Perform PCR
        self.R = self.V.T.dot(np.linalg.inv(S)).dot(U.T).dot(Y)

    def increment(self, X, Y):
        raise NotImplementedError()

    def predict(self, x):
        if self.bias:
            if len(x.shape) == 1:
                x = np.hstack((x, np.ones(1)))
            else:
                x = np.hstack((x, np.ones((x.shape[0], 1))))
        x = np.dot(np.dot(x, self.V.T), self.V)
        return np.dot(x, self.R)


class OptimalLinearRegression(object):
    r"""
    Multivariate Linear Regression using optimal reconstructions.

    Parameters
    ----------
    X : numpy.array
        The regression features used to create the coefficient matrix.
    T : numpy.array
        The shapes differential that denote the dependent variable.
    variance: float or None, Optional
        The SVD variance.
        Default: None

    Raises
    ------
    ValueError
        variance must be set to a number between 0 and 1
    """
    def __init__(self, variance=None, bias=True, eps=1e-10):
        self.variance = variance
        self.bias = bias
        self.R = None
        self.eps = eps

    def train(self, X, Y):
        if self.bias:
            X = np.hstack((X, np.ones((X.shape[0], 1))))

        # Reduce variance of X
        U, l, _ = pca(X, centre=False)
        k = U.shape[0]
        if self.variance is not None:
            variation = np.cumsum(l) / np.sum(l)
            # Inverted for easier parameter semantics
            k = np.sum(variation < self.variance)
            U = U[:k, :]

        # Whitened components
        inv_eig = np.sqrt(np.linalg.inv(np.diag(l[:k])))
        U = inv_eig.dot(U)

        A = X.T.dot(Y).dot(Y.T).dot(X)
        A_tilde = U.dot(A).dot(U.T)

        V, l2, _ = pca(A_tilde, centre=False)
        H = V.dot(U)

        self.R = H.T.dot(np.linalg.pinv(X.dot(H.T)).dot(Y))

    def increment(self, X, Y):
        raise NotImplementedError()

    def predict(self, x):
        if self.bias:
            if len(x.shape) == 1:
                x = np.hstack((x, np.ones(1)))
            else:
                x = np.hstack((x, np.ones((x.shape[0], 1))))
        return np.dot(x, self.R)


class OPPRegression(object):
    r"""
    Multivariate Linear Regression using Orthogonal Procrustes Problem
    reconstructions.

    Parameters
    ----------
    X : numpy.array
        The regression features used to create the coefficient matrix.
    T : numpy.array
        The shapes differential that denote the dependent variable.

    Raises
    ------
    ValueError
        variance must be set to a number between 0 and 1
    """
    def __init__(self, bias=True, eps=1e-10, whiten=False):
        self.bias = bias
        self.R = None
        self.eps = eps
        self.whiten = whiten

    def train(self, X, Y):
        if self.bias:
            # add bias
             X = np.hstack((X, np.ones((X.shape[0], 1))))

        # Whiten
        if self.whiten:
            U, l, _ = pca(X, centre=False)
            inv_eig = np.sqrt(np.linalg.inv(np.diag(l)))
            U = inv_eig.dot(U)
            X = X.dot(U.T).dot(U)

        U, _, V = np.linalg.svd(X.T.dot(Y), full_matrices=False)
        # Skinny SVD
        self.R = U.dot(V)

    def increment(self, X, Y):
        raise NotImplementedError()

    def predict(self, x):
        if self.bias:
            if len(x.shape) == 1:
                x = np.hstack((x, np.ones(1)))
            else:
                x = np.hstack((x, np.ones((x.shape[0], 1))))
        return np.dot(x, self.R)



from __future__ import division
from functools import partial
from menpofit.result import NonParametricAlgorithmResult
import numpy as np
from menpo.visualize import print_dynamic
from menpofit.visualize import print_progress


# TODO: document me!
class BaseSupervisedDescentAlgorithm(object):
    r"""
    """

    def train(self, images, gt_shapes, current_shapes, prefix='',
              verbose=False):
        return self._train(images, gt_shapes, current_shapes, increment=False,
                           prefix=prefix, verbose=verbose)

    def increment(self, images, gt_shapes, current_shapes, prefix='',
                  verbose=False):
        return self._train(images, gt_shapes, current_shapes, increment=True,
                           prefix=prefix, verbose=verbose)

    def _train(self, images, gt_shapes, current_shapes, increment=False,
               prefix='', verbose=False):

        if not increment:
            # Reset the regressors
            self.regressors = []
        elif increment and not (hasattr(self, 'regressors') and self.regressors):
            raise ValueError('Algorithm must be trained before it can be '
                             'incremented.')

        n_perturbations = len(current_shapes[0])
        template_shape = gt_shapes[0]

        # obtain delta_x and gt_x
        delta_x, gt_x = self._compute_delta_x(gt_shapes, current_shapes)

        # Cascaded Regression loop
        for k in range(self.n_iterations):
            # generate regression data
            features_prefix = '{}(Iteration {}) - '.format(prefix, k)
            features = self._compute_training_features(images, gt_shapes,
                                                       current_shapes,
                                                       prefix=features_prefix,
                                                       verbose=verbose)

            if verbose:
                print_dynamic('{}(Iteration {}) - Performing regression'.format(
                    prefix, k))

            if not increment:
                r = self._regressor_cls()
                r.train(features, delta_x)
                self.regressors.append(r)
            else:
                self.regressors[k].increment(features, delta_x)

            # Estimate delta_points
            estimated_delta_x = self.regressors[k].predict(features)
            if verbose:
                self._print_regression_info(template_shape, gt_shapes,
                                            n_perturbations, delta_x,
                                            estimated_delta_x, k,
                                            prefix=prefix)

            self._update_estimates(estimated_delta_x, delta_x, gt_x,
                                   current_shapes)

        return current_shapes

    def _compute_delta_x(self, gt_shapes, current_shapes):
        raise NotImplementedError()

    def _update_estimates(self, estimated_delta_x, delta_x, gt_x,
                          current_shapes):
        raise NotImplementedError()

    def _compute_training_features(self, images, gt_shapes, current_shapes,
                                   prefix='', verbose=False):
        raise NotImplementedError()

    def _compute_test_features(self, image, current_shape):
        raise NotImplementedError()

    def run(self, image, initial_shape, gt_shape=None, **kwargs):
        raise NotImplementedError()

    def _print_regression_info(self, template_shape, gt_shapes, n_perturbations,
                               delta_x, estimated_delta_x, level_index,
                               prefix=''):
        raise NotImplementedError()



# TODO: document me!
def features_per_patch(image, shape, patch_shape, features_callable):
    """r
    """
    patches = image.extract_patches(shape, patch_shape=patch_shape,
                                    as_single_array=True)

    patch_features = [features_callable(p[0]).ravel() for p in patches]
    return np.hstack(patch_features)


# TODO: document me!
def features_per_shapes(image, shapes, patch_shape, features_callable):
    """r
    """
    patch_features = [features_per_patch(image, s, patch_shape,
                                         features_callable)
                      for s in shapes]

    return np.vstack(patch_features)


# TODO: document me!
def features_per_image(images, shapes, patch_shape, features_callable,
                       prefix='', verbose=False):
    """r
    """
    wrap = partial(print_progress,
                   prefix='{}Extracting patches'.format(prefix),
                   end_with_newline=not prefix, verbose=verbose)

    patch_features = [features_per_shapes(i, shapes[j], patch_shape,
                                          features_callable)
                      for j, i in enumerate(wrap(images))]
    return np.vstack(patch_features)


def compute_non_parametric_delta_x(gt_shapes, current_shapes):
    r"""
    """
    n_x = gt_shapes[0].n_parameters
    n_gt_shapes = len(gt_shapes)
    n_current_shapes = len(current_shapes[0])

    # initialize current, ground truth and delta parameters
    gt_x = np.empty((n_gt_shapes * n_current_shapes, n_x))
    delta_x = np.empty((n_gt_shapes * n_current_shapes, n_x))

    # obtain ground truth points and compute delta points
    k = 0
    for gt_s, shapes in zip(gt_shapes, current_shapes):
        c_gt_s = gt_s.as_vector()
        for s in shapes:
            # compute ground truth shape vector
            gt_x[k] = c_gt_s
            # compute delta shape vector
            delta_x[k] = c_gt_s - s.as_vector()
            k += 1

    return delta_x, gt_x


def update_non_parametric_estimates(estimated_delta_x, delta_x, gt_x,
                                    current_shapes):
    j = 0
    for shapes in current_shapes:
        for s in shapes:
            # update current x
            current_x = s.as_vector() + estimated_delta_x[j]
            # update current shape inplace
            s._from_vector_inplace(current_x)
            # update delta_x
            delta_x[j] = gt_x[j] - current_x
            # increase index
            j += 1


def print_non_parametric_info(template_shape, gt_shapes, n_perturbations,
                              delta_x, estimated_delta_x, level_index,
                              compute_error_f, prefix=''):
    print_dynamic('{}(Iteration {}) - Calculating errors'.format(
        prefix, level_index))
    errors = []
    for j, (dx, edx) in enumerate(zip(delta_x, estimated_delta_x)):
        s1 = template_shape.from_vector(dx)
        s2 = template_shape.from_vector(edx)
        gt_s = gt_shapes[np.floor_divide(j, n_perturbations)]
        errors.append(compute_error_f(s1, s2, gt_s))
    mean = np.mean(errors)
    std = np.std(errors)
    median = np.median(errors)
    print_dynamic('{}(Iteration {}) - Training error -> '
                  'mean: {:.4f}, std: {:.4f}, median: {:.4f}.\n'.
                  format(prefix, level_index, mean, std, median))


def print_parametric_info(model, gt_shapes, n_perturbations,
                          delta_x, estimated_delta_x, level_index,
                          compute_error_f, prefix=''):
    print_dynamic('{}(Iteration {}) - Calculating errors'.format(
        prefix, level_index))
    errors = []
    for j, (dx, edx) in enumerate(zip(delta_x, estimated_delta_x)):
        model._from_vector_inplace(dx)
        s1 = model.target
        model._from_vector_inplace(edx)
        s2 = model.target

        gt_s = gt_shapes[np.floor_divide(j, n_perturbations)]
        errors.append(compute_error_f(s1, s2, gt_s))
    mean = np.mean(errors)
    std = np.std(errors)
    median = np.median(errors)
    print_dynamic('{}(Iteration {}) - Training error -> '
                  'mean: {:.4f}, std: {:.4f}, median: {:.4f}.\n'.
                  format(prefix, level_index, mean, std, median))


def compute_parametric_delta_x(gt_shapes, current_shapes, model):
    # initialize current and delta parameters arrays
    n_samples = len(gt_shapes) * len(current_shapes[0])
    gt_params = np.empty((n_samples, model.n_parameters))
    delta_params = np.empty_like(gt_params)

    k = 0
    for gt_s, c_s in zip(gt_shapes, current_shapes):
        # Compute and cache ground truth parameters
        model.set_target(gt_s)
        c_gt_params = model.as_vector()
        for s in c_s:
            gt_params[k] = c_gt_params

            model.set_target(s)
            current_params = model.as_vector()
            delta_params[k] = c_gt_params - current_params

            k += 1

    return delta_params, gt_params


def update_parametric_estimates(estimated_delta_x, delta_x, gt_x,
                                current_shapes, model):
    j = 0
    for shapes in current_shapes:
        for s in shapes:
            # Estimate parameters
            edx = estimated_delta_x[j]
            # Current parameters
            model.set_target(s)
            cx = model.as_vector() + edx
            model._from_vector_inplace(cx)

            # Update current shape inplace
            s._from_vector_inplace(model.target.as_vector().copy())

            delta_x[j] = gt_x[j] - cx
            j += 1


def build_appearance_model(images, gt_shapes, patch_shape, patch_features,
                           appearance_model_cls, verbose=False, prefix=''):
    wrap = partial(print_progress,
                   prefix='{}Extracting ground truth patches'.format(prefix),
                   end_with_newline=not prefix, verbose=verbose)
    n_images = len(images)
    # Extract patches from ground truth
    gt_patches = [features_per_patch(im, gt_s, patch_shape,
                                     patch_features)
                  for gt_s, im in wrap(zip(gt_shapes, images))]
    # Calculate appearance model from extracted gt patches
    gt_patches = np.array(gt_patches).reshape([n_images, -1])
    if verbose:
        print_dynamic('{}Building Appearance Model'.format(prefix))
    return appearance_model_cls(gt_patches)


def fit_parametric_shape(image, initial_shape, parametric_algo, gt_shape=None):
    # set current shape and initialize list of shapes
    parametric_algo.shape_model.set_target(initial_shape)
    current_shape = initial_shape.from_vector(
        parametric_algo.shape_model.target.as_vector().copy())
    shapes = [current_shape]

    # Cascaded Regression loop
    for r in parametric_algo.regressors:
        # compute regression features
        features = parametric_algo._compute_test_features(image, current_shape)

        # solve for increments on the shape vector
        dx = r.predict(features).ravel()

        # update current shape
        p = parametric_algo.shape_model.as_vector() + dx
        parametric_algo.shape_model._from_vector_inplace(p)
        current_shape = current_shape.from_vector(
            parametric_algo.shape_model.target.as_vector().copy())
        shapes.append(current_shape)

    # return algorithm result
    return NonParametricAlgorithmResult(image, shapes,
                                        gt_shape=gt_shape)


def fit_non_parametric_shape(image, initial_shape, non_parametric_algo,
                             gt_shape=None):
    # set current shape and initialize list of shapes
    current_shape = initial_shape
    shapes = [initial_shape]

    # Cascaded Regression loop
    for r in non_parametric_algo.regressors:
        # compute regression features
        features = non_parametric_algo._compute_test_features(image,
                                                              current_shape)

        # solve for increments on the shape vector
        dx = r.predict(features)

        # update current shape
        current_shape = current_shape.from_vector(
            current_shape.as_vector() + dx)
        shapes.append(current_shape)

    # return algorithm result
    return NonParametricAlgorithmResult(image, shapes,
                                        gt_shape=gt_shape)



from __future__ import division
import numpy as np
from menpofit.aam.algorithm.lk import (LucasKanadeBaseInterface,
                                       LucasKanadePatchBaseInterface)
from .result import ATMAlgorithmResult, LinearATMAlgorithmResult


# TODO document me!
class ATMLKStandardInterface(LucasKanadeBaseInterface):
    r"""
    """
    def __init__(self, transform, template, sampling=None):
        super(ATMLKStandardInterface, self).__init__(transform, template,
                                                     sampling=sampling)

    def algorithm_result(self, image, shape_parameters, cost_functions=None,
                         gt_shape=None):
        return ATMAlgorithmResult(
            image, self, shape_parameters,
            cost_functions=cost_functions, gt_shape=gt_shape)


# TODO document me!
class ATMLKLinearInterface(ATMLKStandardInterface):
    r"""
    """
    @property
    def shape_model(self):
        return self.transform.model

    def algorithm_result(self, image, shape_parameters, cost_functions=None,
                         gt_shape=None):
        return LinearATMAlgorithmResult(
            image, self, shape_parameters,
            cost_functions=cost_functions, gt_shape=gt_shape)


# TODO document me!
class ATMLKPatchInterface(LucasKanadePatchBaseInterface):
    r"""
    """
    def algorithm_result(self, image, shape_parameters, cost_functions=None,
                         gt_shape=None):
        return ATMAlgorithmResult(
            image, self, shape_parameters,
            cost_functions=cost_functions, gt_shape=gt_shape)


# TODO document me!
class LucasKanade(object):

    def __init__(self, atm_interface, eps=10**-5):
        self.eps = eps
        self.interface = atm_interface
        self._precompute()

    @property
    def transform(self):
        return self.interface.transform

    @property
    def template(self):
        return self.interface.template

    def _precompute(self):
        # grab number of shape and appearance parameters
        self.n = self.transform.n_parameters

        # vectorize template and mask it
        self.t_m = self.template.as_vector()[self.interface.i_mask]

        # compute warp jacobian
        self.dW_dp = self.interface.warp_jacobian()

        # compute shape model prior
        # TODO: Is this correct? It's like modelling no noise at all
        noise_variance = self.interface.shape_model.noise_variance() or 1
        s2 = 1.0 / noise_variance
        L = self.interface.shape_model.eigenvalues
        self.s2_inv_L = np.hstack((np.ones((4,)), s2 / L))


# TODO document me!
class Compositional(LucasKanade):
    r"""
    Abstract Interface for Compositional ATM algorithms
    """
    def run(self, image, initial_shape, gt_shape=None, max_iters=20,
            map_inference=False):
        # define cost closure
        def cost_closure(x):
            return lambda: x.T.dot(x)

        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Compositional Gauss-Newton loop -------------------------------------

        # warp image
        self.i = self.interface.warp(image)
        # vectorize it and mask it
        i_m = self.i.as_vector()[self.interface.i_mask]

        # compute masked error
        self.e_m = i_m - self.t_m

        # update cost
        cost_functions = [cost_closure(self.e_m)]

        while k < max_iters and eps > self.eps:
            # solve for increments on the shape parameters
            self.dp = self._solve(map_inference)

            # update warp
            s_k = self.transform.target.points
            self._update_warp()
            p_list.append(self.transform.as_vector())

            # warp image
            self.i = self.interface.warp(image)
            # vectorize it and mask it
            i_m = self.i.as_vector()[self.interface.i_mask]

            # compute masked error
            self.e_m = i_m - self.t_m

            # update cost
            cost_functions.append(cost_closure(self.e_m))

            # test convergence
            eps = np.abs(np.linalg.norm(s_k - self.transform.target.points))

            # increase iteration counter
            k += 1

        # return algorithm result
        return self.interface.algorithm_result(
            image, p_list, cost_functions=cost_functions, gt_shape=gt_shape)


# TODO document me!
class ForwardCompositional(Compositional):
    r"""
    Forward Compositional (FC) Gauss-Newton algorithm
    """
    def _solve(self, map_inference):
        # compute warped image gradient
        nabla_i = self.interface.gradient(self.i)
        # compute masked forward Jacobian
        J_m = self.interface.steepest_descent_images(nabla_i, self.dW_dp)
        # compute masked forward Hessian
        JJ_m = J_m.T.dot(J_m)
        # solve for increments on the shape parameters
        if map_inference:
            return self.interface.solve_shape_map(
                JJ_m, J_m, self.e_m,  self.s2_inv_L,
                self.transform.as_vector())
        else:
            return self.interface.solve_shape_ml(JJ_m, J_m, self.e_m)

    def _update_warp(self):
        # update warp based on forward composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() + self.dp)


# TODO document me!
class InverseCompositional(Compositional):
    r"""
    Inverse Compositional (IC) Gauss-Newton algorithm
    """
    def _precompute(self):
        # call super method
        super(InverseCompositional, self)._precompute()
        # compute appearance model mean gradient
        nabla_t = self.interface.gradient(self.template)
        # compute masked inverse Jacobian
        self.J_m = self.interface.steepest_descent_images(-nabla_t, self.dW_dp)
        # compute masked inverse Hessian
        self.JJ_m = self.J_m.T.dot(self.J_m)
        # compute masked Jacobian pseudo-inverse
        self.pinv_J_m = np.linalg.solve(self.JJ_m, self.J_m.T)

    def _solve(self, map_inference):
        # solve for increments on the shape parameters
        if map_inference:
            return self.interface.solve_shape_map(
                self.JJ_m, self.J_m, self.e_m, self.s2_inv_L,
                self.transform.as_vector())
        else:
            return -self.pinv_J_m.dot(self.e_m)

    def _update_warp(self):
        # update warp based on inverse composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() - self.dp)



import dlib
import numpy as np
from menpo.shape import PointCloud, bounding_box


all_parts = lambda det: ((p.y, p.x) for p in det.parts())


def pointcloud_to_dlib_parts(pcloud):
    return [dlib.point(int(p[1]), int(p[0])) for p in pcloud.points]


def dlib_full_object_detection_to_pointcloud(full_object_detection):
    return PointCloud(np.vstack(all_parts(full_object_detection)), copy=False)


def dlib_rect_to_bounding_box(rect):
    return bounding_box((rect.top(), rect.left()),
                        (rect.bottom(), rect.right()))


def pointcloud_to_dlib_rect(pg):
    min_p, max_p = pg.bounds()
    return dlib.rectangle(left=int(min_p[1]), top=int(min_p[0]),
                          right=int(max_p[1]), bottom=int(max_p[0]))


def bounding_box_pointcloud_to_dlib_fo_detection(bbox, pcloud):
    return dlib.full_object_detection(
        pointcloud_to_dlib_rect(bbox.bounding_box()),
        pointcloud_to_dlib_parts(pcloud))


def copy_dlib_options(options):
    new_options = dlib.shape_predictor_training_options()
    for p in sorted(filter(lambda x: '__' not in x, dir(options))):
        setattr(new_options, p, getattr(options, p))
    return new_options


def image_to_dlib_pixels(im):
    pixels = np.array(im.as_PILImage())
    # Only supports RGB and Grayscale
    if im.n_channels > 3:
        pixels = pixels[..., 0]
    return pixels



from menpofit.result import ParametricAlgorithmResult, MultiFitterResult


# TODO: document me!
class CLMAlgorithmResult(ParametricAlgorithmResult):
    r"""
    """


# TODO: document me!
class CLMFitterResult(MultiFitterResult):
    r"""
    """




from __future__ import division
import numpy as np
import warnings
from menpo.feature import ndfeature


# TODO: Document me!
@ndfeature
def centralize(x, axes=(-2, -1)):
    r"""
    """
    mean = np.mean(x, axis=axes, keepdims=True)
    return x - mean


# TODO: Document me!
@ndfeature
def normalize_norm(x, axes=(-2, -1)):
    r"""
    """
    x = centralize(x, axes=axes)
    norm = np.asarray(np.linalg.norm(x, axis=axes))
    positions = np.asarray(axes) + len(x.shape)
    for axis in positions:
        norm = np.expand_dims(norm, axis=axis)
    return handle_div_by_zero(x, norm)


# TODO: document me!
@ndfeature
def normalize_std(x, axes=(-2, -1)):
    r"""
    """
    x = centralize(x, axes=axes)
    std = np.std(x, axis=axes, keepdims=True)
    return handle_div_by_zero(x, std)


# TODO: document me!
@ndfeature
def normalize_var(x, axes=(-2, -1)):
    r"""
    """
    x = centralize(x, axes=axes)
    var = np.var(x, axis=axes, keepdims=True)
    return handle_div_by_zero(x, var)


# TODO: document me!
@ndfeature
def probability_map(x, axes=(-2, -1)):
    r"""
    """
    x = x - np.min(x, axis=axes, keepdims=True)
    total = np.sum(x, axis=axes, keepdims=True)
    nonzero = total > 0
    if np.any(~nonzero):
        warnings.warn("some of x axes have 0 variance - uniform probability "
                      "maps are used them.")
        x[nonzero] /= total[nonzero]
        x[~nonzero] = 1 / np.prod(axes)
    else:
        x /= total
    return x


# TODO: document me!
def handle_div_by_zero(x, normalizer):
    r"""
    """
    nonzero = normalizer > 0
    if np.any(~nonzero):
        warnings.warn("some of the denominators have 0 variance - they cannot "
                      "be normalized.")
        x[nonzero] /= normalizer[nonzero]
    else:
        x /= normalizer
    return x



from __future__ import division
import warnings
import numpy as np
from functools import wraps
from menpo.feature.base import rebuild_feature_image
try:
    # try importing pyfftw
    from pyfftw.interfaces.numpy_fft import fft2, ifft2, fftshift, ifftshift

    try:
        # try calling fft2 on a 4-dimensional array (this is known to have
        # problem in some linux distributions)
        fft2(np.zeros((1, 1, 1, 1)))
    except RuntimeError:
        warnings.warn("pyfftw is known to be buggy on your system, numpy.fft "
                      "will be used instead. Consequently, all algorithms "
                      "using ffts will be running at a slower speed.",
                      RuntimeWarning)
        from numpy.fft import fft2, ifft2, fftshift, ifftshift
except ImportError:
    warnings.warn("pyfftw is not installed on your system, numpy.fft will be "
                  "used instead. Consequently, all algorithms using ffts "
                  "will be running at a slower speed. Consider installing "
                  "pyfftw (pip install pyfftw) to speed up your ffts.",
                  ImportWarning)
    from numpy.fft import fft2, ifft2, fftshift, ifftshift


# TODO: Document me!
def pad(pixels, ext_shape, boundary='constant'):
    r"""
    """
    h, w = pixels.shape[-2:]

    h_margin = (ext_shape[0] - h) // 2
    w_margin = (ext_shape[1] - w) // 2

    h_margin2 = h_margin
    if h + 2 * h_margin < ext_shape[0]:
        h_margin += 1

    w_margin2 = w_margin
    if w + 2 * w_margin < ext_shape[1]:
        w_margin += 1

    pad_width = []
    for _ in pixels.shape[:-2]:
        pad_width.append((0, 0))
    pad_width += [(h_margin, h_margin2), (w_margin, w_margin2)]
    pad_width = tuple(pad_width)

    return np.lib.pad(pixels, pad_width, mode=boundary)


# TODO: Document me!
def crop(pixels, shape):
    r"""
    """
    h, w = pixels.shape[-2:]

    h_margin = (h - shape[0]) // 2
    w_margin = (w - shape[1]) // 2

    h_corrector = 1 if np.remainder(h - shape[0], 2) != 0 else 0
    w_corrector = 1 if np.remainder(w - shape[1], 2) != 0 else 0

    return pixels[...,
                  h_margin + h_corrector:-h_margin,
                  w_margin + w_corrector:-w_margin]


# TODO: Document me!
def ndconvolution(wrapped):
    r"""
    """
    @wraps(wrapped)
    def wrapper(image, filter, *args, **kwargs):
        if not isinstance(image, np.ndarray) and not isinstance(filter, np.ndarray):
            # Both image and filter are menpo images
            feature = wrapped(image.pixels, filter.pixels, *args, **kwargs)
            return rebuild_feature_image(image, feature)
        elif not isinstance(image, np.ndarray):
            # Image is menpo image
            feature = wrapped(image.pixels, filter, *args, **kwargs)
            return rebuild_feature_image(image, feature)
        elif not isinstance(filter, np.ndarray):
            # filter is menpo image
            return wrapped(image, filter, *args, **kwargs)
        else:
            return wrapped(image, filter, *args, **kwargs)
    return wrapper


# TODO: Document me!
@ndconvolution
def fft_convolve2d(x, f, mode='same', boundary='constant', fft_filter=False):
    r"""
    Performs fast 2d convolution in the frequency domain convolving each image
    channel with its corresponding filter channel.

    Parameters
    ----------
    x : ``(channels, height, width)`` `ndarray`
        Image.
    f : ``(channels, height, width)`` `ndarray`
        Filter.
    mode : str {`full`, `same`, `valid`}, optional
        Determines the shape of the resulting convolution.
    boundary: str {`constant`, `symmetric`}, optional
        Determines how the image is padded.
    fft_filter: `bool`, optional
        If `True`, the filter is assumed to be defined on the frequency
        domain. If `False` the filter is assumed to be defined on the
        spatial domain.

    Returns
    -------
    c: ``(channels, height, width)`` `ndarray`
        Result of convolving each image channel with its corresponding
        filter channel.
    """
    if fft_filter:
        # extended shape is filter shape
        ext_shape = np.asarray(f.shape[-2:])

        # extend image and filter
        ext_x = pad(x, ext_shape, boundary=boundary)

        # compute ffts of extended image
        fft_ext_x = fft2(ext_x)
        fft_ext_f = f
    else:
        # extended shape
        x_shape = np.asarray(x.shape[-2:])
        f_shape = np.asarray(f.shape[-2:])
        f_half_shape = (f_shape / 2).astype(int)
        ext_shape = x_shape + f_half_shape - 1

        # extend image and filter
        ext_x = pad(x, ext_shape, boundary=boundary)
        ext_f = pad(f, ext_shape)

        # compute ffts of extended image and extended filter
        fft_ext_x = fft2(ext_x)
        fft_ext_f = fft2(ext_f)

    # compute extended convolution in Fourier domain
    fft_ext_c = fft_ext_f * fft_ext_x

    # compute ifft of extended convolution
    ext_c = np.real(ifftshift(ifft2(fft_ext_c), axes=(-2, -1)))

    if mode is 'full':
        return ext_c
    elif mode is 'same':
        return crop(ext_c, x_shape)
    elif mode is 'valid':
        return crop(ext_c, x_shape - f_half_shape + 1)
    else:
        raise ValueError(
            "mode={}, is not supported. The only supported "
            "modes are: 'full', 'same' and 'valid'.".format(mode))


# TODO: Document me!
@ndconvolution
def fft_convolve2d_sum(x, f, mode='same', boundary='constant',
                       fft_filter=False, axis=0, keepdims=True):
    r"""
    Performs fast 2d convolution in the frequency domain convolving each image
    channel with its corresponding filter channel and summing across the
    channel axis.

    Parameters
    ----------
    x : ``(channels, height, width)`` `ndarray`
        Image.
    f : ``(channels, height, width)`` `ndarray`
        Filter.
    mode : str {`full`, `same`, `valid`}, optional
        Determines the shape of the resulting convolution.
    boundary: str {`constant`, `symmetric`}, optional
        Determines how the image is padded.
    fft_filter: `bool`, optional
        If `True`, the filter is assumed to be defined on the frequency
        domain. If `False` the filter is assumed to be defined on the
        spatial domain.
    axis : `int`, optional
        The axis across to which the summation is performed.
    keepdims: `boolean`, optional
        If `True` the number of dimensions of the result is the same as the
        number of dimensions of the filter. If `False` the channel dimension
        is lost in the result.
    Returns
    -------
    c: ``(1, height, width)`` `ndarray`
        Result of convolving each image channel with its corresponding
        filter channel and summing across the channel axis.
    """
    if fft_filter:
        fft_ext_f = f

        # extended shape is fft_ext_filter shape
        x_shape = np.asarray(x.shape[-2:])
        f_shape = ((np.asarray(fft_ext_f.shape[-2:]) + 1) / 1.5).astype(int)
        f_half_shape = (f_shape / 2).astype(int)
        ext_shape = np.asarray(f.shape[-2:])

        # extend image and filter
        ext_x = pad(x, ext_shape, boundary=boundary)

        # compute ffts of extended image
        fft_ext_x = fft2(ext_x)
    else:
        # extended shape
        x_shape = np.asarray(x.shape[-2:])
        f_shape = np.asarray(f.shape[-2:])
        f_half_shape = (f_shape / 2).astype(int)
        ext_shape = x_shape + f_half_shape - 1

        # extend image and filter
        ext_x = pad(x, ext_shape, boundary=boundary)
        ext_f = pad(f, ext_shape)

        # compute ffts of extended image and extended filter
        fft_ext_x = fft2(ext_x)
        fft_ext_f = fft2(ext_f)

    # compute extended convolution in Fourier domain
    fft_ext_c = np.sum(fft_ext_f * fft_ext_x, axis=axis, keepdims=keepdims)

    # compute ifft of extended convolution
    ext_c = np.real(ifftshift(ifft2(fft_ext_c), axes=(-2, -1)))

    if mode is 'full':
        return ext_c
    elif mode is 'same':
        return crop(ext_c, x_shape)
    elif mode is 'valid':
        return crop(ext_c, x_shape - f_half_shape + 1)
    else:
        raise ValueError(
            "mode={}, is not supported. The only supported "
            "modes are: 'full', 'same' and 'valid'.".format(mode))



from __future__ import division
import numpy as np
from functools import partial
import warnings

from menpo.transform import Scale
from menpo.feature import no_op
from menpo.base import name_of_callable

from menpofit.visualize import print_progress
from menpofit.base import batch
from menpofit.builder import (scale_images, rescale_images_to_reference_shape,
                              compute_reference_shape, MenpoFitBuilderWarning,
                              compute_features)
from menpofit.fitter import (MultiFitter, noisy_shape_from_bounding_box,
                             align_shape_with_bounding_box,
                             generate_perturbations_from_gt)
from menpofit.result import MultiFitterResult
import menpofit.checks as checks


# TODO: document me!
class SupervisedDescentFitter(MultiFitter):
    r"""
    """
    def __init__(self, images, group=None, bounding_box_group_glob=None,
                 reference_shape=None, sd_algorithm_cls=None,
                 holistic_features=no_op, patch_features=no_op,
                 patch_shape=(17, 17), diagonal=None, scales=(0.5, 1.0),
                 n_iterations=3, n_perturbations=30,
                 perturb_from_gt_bounding_box=noisy_shape_from_bounding_box,
                 batch_size=None, verbose=False):

        if batch_size is not None:
            raise NotImplementedError('Training an SDM with a batch size '
                                      '(incrementally) is not implemented yet.')
        # check parameters
        checks.check_diagonal(diagonal)
        scales = checks.check_scales(scales)
        n_scales = len(scales)
        patch_features = checks.check_callable(patch_features, n_scales)
        sd_algorithm_cls = checks.check_callable(sd_algorithm_cls, n_scales)
        holistic_features = checks.check_callable(holistic_features, n_scales)
        patch_shape = checks.check_patch_shape(patch_shape, n_scales)
        # set parameters
        self.algorithms = []
        self.reference_shape = reference_shape
        self._sd_algorithm_cls = sd_algorithm_cls
        self.holistic_features = holistic_features
        self.patch_features = patch_features
        self.patch_shape = patch_shape
        self.diagonal = diagonal
        self.scales = scales
        self.n_perturbations = n_perturbations
        self.n_iterations = checks.check_max_iters(n_iterations, n_scales)
        self._perturb_from_gt_bounding_box = perturb_from_gt_bounding_box
        # set up algorithms
        self._setup_algorithms()

        # Now, train the model!
        self._train(images, increment=False,  group=group,
                    bounding_box_group_glob=bounding_box_group_glob,
                    verbose=verbose, batch_size=batch_size)

    def _setup_algorithms(self):
        for j in range(self.n_scales):
            self.algorithms.append(self._sd_algorithm_cls[j](
                patch_features=self.patch_features[j],
                patch_shape=self.patch_shape[j],
                n_iterations=self.n_iterations[j]))

    def _train(self, images, increment=False, group=None,
               bounding_box_group_glob=None, verbose=False, batch_size=None):
        r"""
        """
        # If batch_size is not None, then we may have a generator, else we
        # assume we have a list.
        # If batch_size is not None, then we may have a generator, else we
        # assume we have a list.
        if batch_size is not None:
            # Create a generator of fixed sized batches. Will still work even
            # on an infinite list.
            image_batches = batch(images, batch_size)
        else:
            image_batches = [list(images)]

        for k, image_batch in enumerate(image_batches):
            if k == 0:
                if self.reference_shape is None:
                    # If no reference shape was given, use the mean of the first
                    # batch
                    if batch_size is not None:
                        warnings.warn('No reference shape was provided. The '
                                      'mean of the first batch will be the '
                                      'reference shape. If the batch mean is '
                                      'not representative of the true mean, '
                                      'this may cause issues.',
                                      MenpoFitBuilderWarning)
                    self.reference_shape = compute_reference_shape(
                        [i.landmarks[group].lms for i in image_batch],
                        self.diagonal, verbose=verbose)
            # We set landmarks on the images to archive the perturbations, so
            # when the default 'None' is used, we need to grab the actual
            # label to sort out the ambiguity
            if group is None:
                group = image_batch[0].landmarks.group_labels[0]

            # After the first batch, we are incrementing the model
            if k > 0:
                increment = True

            if verbose:
                print('Computing batch {}'.format(k))

            # Train each batch
            self._train_batch(
                image_batch, increment=increment, group=group,
                bounding_box_group_glob=bounding_box_group_glob,
                verbose=verbose)

    def _train_batch(self, image_batch, increment=False, group=None,
                     bounding_box_group_glob=None, verbose=False):
        # Rescale to existing reference shape
        image_batch = rescale_images_to_reference_shape(
            image_batch, group, self.reference_shape,
            verbose=verbose)

        generated_bb_func = generate_perturbations_from_gt(
            image_batch, self.n_perturbations,
            self._perturb_from_gt_bounding_box, gt_group=group,
            bb_group_glob=bounding_box_group_glob, verbose=verbose)

        # for each scale (low --> high)
        current_shapes = []
        for j in range(self.n_scales):
            if verbose:
                if len(self.scales) > 1:
                    scale_prefix = '  - Scale {}: '.format(j)
                else:
                    scale_prefix = '  - '
            else:
                scale_prefix = None

            # Handle holistic features
            if j == 0 and self.holistic_features[j] == no_op:
                # Saves a lot of memory
                feature_images = image_batch
            elif j == 0 or self.holistic_features[j] is not self.holistic_features[j - 1]:
                # Compute features only if this is the first pass through
                # the loop or the features at this scale are different from
                # the features at the previous scale
                feature_images = compute_features(image_batch,
                                                  self.holistic_features[j],
                                                  prefix=scale_prefix,
                                                  verbose=verbose)
            # handle scales
            if self.scales[j] != 1:
                # Scale feature images only if scale is different than 1
                scaled_images = scale_images(feature_images, self.scales[j],
                                             prefix=scale_prefix,
                                             verbose=verbose)
            else:
                scaled_images = feature_images

            # Extract scaled ground truth shapes for current scale
            scaled_shapes = [i.landmarks[group].lms for i in scaled_images]

            if j == 0:
                msg = '{}Aligning reference shape with bounding boxes.'.format(
                    scale_prefix)
                wrap = partial(print_progress, prefix=msg,
                               end_with_newline=False, verbose=verbose)

                # Extract perturbations at the very bottom level
                for ii in wrap(scaled_images):
                    c_shapes = []
                    for bbox in generated_bb_func(ii):
                        c_s = align_shape_with_bounding_box(
                            self.reference_shape, bbox)
                        c_shapes.append(c_s)
                    current_shapes.append(c_shapes)

            # train supervised descent algorithm
            if not increment:
                current_shapes = self.algorithms[j].train(
                    scaled_images, scaled_shapes, current_shapes,
                    prefix=scale_prefix, verbose=verbose)
            else:
                current_shapes = self.algorithms[j].increment(
                    scaled_images, scaled_shapes, current_shapes,
                    prefix=scale_prefix, verbose=verbose)

            # Scale current shapes to next resolution, don't bother
            # scaling final level
            if j != (self.n_scales - 1):
                transform = Scale(self.scales[j + 1] / self.scales[j],
                                  n_dims=2)
                for image_shapes in current_shapes:
                    for k, shape in enumerate(image_shapes):
                        image_shapes[k] = transform.apply(shape)

    def increment(self, images, group=None, bounding_box_group=None,
                  verbose=False, batch_size=None):
        raise NotImplementedError('Incrementing SDM methods is not yet '
                                  'implemented as careful attention must '
                                  'be taken when considering the relationships '
                                  'between cascade levels.')

    def perturb_from_bb(self, gt_shape, bb):
        return self._perturb_from_gt_bounding_box(gt_shape, bb)

    def perturb_from_gt_bb(self, gt_bb):
        return self._perturb_from_gt_bounding_box(gt_bb, gt_bb)

    def _fitter_result(self, image, algorithm_results, affine_correction,
                       gt_shape=None):
        return MultiFitterResult(image, self, algorithm_results,
                                 affine_correction, gt_shape=gt_shape)

    def __str__(self):
        if self.diagonal is not None:
            diagonal = self.diagonal
        else:
            y, x = self.reference_shape.range()
            diagonal = np.sqrt(x ** 2 + y ** 2)
        is_custom_perturb_func = (self._perturb_from_gt_bounding_box !=
                                  noisy_shape_from_bounding_box)
        regressor_cls = self.algorithms[0]._regressor_cls

        # Compute scale info strings
        scales_info = []
        lvl_str_tmplt = r"""  - Scale {}
   - {} iterations
   - Patch shape: {}
   - Holistic feature: {}
   - Patch feature: {}"""
        for k, s in enumerate(self.scales):
            scales_info.append(lvl_str_tmplt.format(
                s, self.n_iterations[k], self.patch_shape[k],
                name_of_callable(self.holistic_features[k]),
                name_of_callable(self.patch_features[k])))
        scales_info = '\n'.join(scales_info)

        cls_str = r"""Supervised Descent Method
 - Regression performed using the {reg_alg} algorithm
   - Regression class: {reg_cls}
 - Perturbations generated per shape: {n_perturbations}
 - Images scaled to diagonal: {diagonal:.2f}
 - Custom perturbation scheme used: {is_custom_perturb_func}
 - Scales: {scales}
{scales_info}
""".format(
            reg_alg=name_of_callable(self._sd_algorithm_cls),
            reg_cls=name_of_callable(regressor_cls),
            n_perturbations=self.n_perturbations,
            diagonal=diagonal,
            is_custom_perturb_func=is_custom_perturb_func,
            scales=self.scales,
            scales_info=scales_info)
        return cls_str

# *
# ************************* Non-Parametric Fitters *****************************
# *
from .algorithm import NonParametricNewton

# Aliases for common combinations of supervised descent fitting
SDM = partial(SupervisedDescentFitter, sd_algorithm_cls=NonParametricNewton)


class RegularizedSDM(SupervisedDescentFitter):
    def __init__(self, images, group=None, bounding_box_group_glob=None,
                 alpha=0.0001, reference_shape=None,
                 holistic_features=no_op, patch_features=no_op,
                 patch_shape=(17, 17), diagonal=None, scales=(0.5, 1.0),
                 n_iterations=6, n_perturbations=30,
                 perturb_from_gt_bounding_box=noisy_shape_from_bounding_box,
                 batch_size=None, verbose=False):
        super(RegularizedSDM, self).__init__(
            images, group=group,
            bounding_box_group_glob=bounding_box_group_glob,
            reference_shape=reference_shape,
            sd_algorithm_cls=partial(NonParametricNewton, alpha=alpha),
            holistic_features=holistic_features, patch_features=patch_features,
            patch_shape=patch_shape, diagonal=diagonal, scales=scales,
            n_iterations=n_iterations, n_perturbations=n_perturbations,
            perturb_from_gt_bounding_box=perturb_from_gt_bounding_box,
            batch_size=batch_size, verbose=verbose)



from menpofit import checks
from menpofit.fitter import ModelFitter
from menpofit.modelinstance import OrthoPDM
from .algorithm import (
    GradientDescentCLMAlgorithm, RegularisedLandmarkMeanShift)
from .result import CLMFitterResult


# TODO: Document me!
class CLMFitter(ModelFitter):
    r"""
    """
    @property
    def clm(self):
        return self._model

    def _fitter_result(self, image, algorithm_results, affine_correction,
                       gt_shape=None):
        return CLMFitterResult(image, self, algorithm_results,
                               affine_correction, gt_shape=gt_shape)


# TODO: Document me!
# TODO: Rethink shape model and OrthoPDM relation
class GradientDescentCLMFitter(CLMFitter):
    r"""
    """
    def __init__(self, clm, gd_algorithm_cls=RegularisedLandmarkMeanShift,
                 n_shape=None):
        self._model = clm
        self._gd_algorithms_cls = checks.check_algorithm_cls(
            gd_algorithm_cls, self.n_scales, GradientDescentCLMAlgorithm)
        self._check_n_shape(n_shape)

        self.algorithms = []
        for i in range(self.clm.n_scales):
            algorithm = self._gd_algorithms_cls[i](
                self.clm.expert_ensembles[i],
                self.clm.shape_models[i])
            self.algorithms.append(algorithm)


# TODO: Implement me!
# TODO: Document me!
class SupervisedDescentCLMFitter(CLMFitter):
    r"""
    """
    def __init__(self):
        raise NotImplementedError()



from .base import CLM
from .fitter import GradientDescentCLMFitter
from .algorithm import ActiveShapeModel, RegularisedLandmarkMeanShift
from .expert import (
    CorrelationFilterExpertEnsemble, IncrementalCorrelationFilterThinWrapper)



import numpy as np
from menpofit.math.correlationfilter import mccf, imccf


# TODO: document me!
class IncrementalCorrelationFilterThinWrapper(object):
    r"""
    """
    def __init__(self, cf_callable=mccf, icf_callable=imccf):
        self.cf_callable = cf_callable
        self.icf_callable = icf_callable

    def increment(self, A, B, n_x, Z, t):
        r"""
        """
        # Turn list of X into ndarray
        if isinstance(Z, list):
            Z = np.asarray(Z)
        return self.icf_callable(A, B, n_x, Z, t)

    def train(self, X, t):
        r"""
        """
        # Turn list of X into ndarray
        if isinstance(X, list):
            X = np.asarray(X)
        # Return linear svm filter and bias
        return self.cf_callable(X, t)



import numpy as np
from numpy.fft import fft2, ifft2, ifftshift
from scipy.sparse import spdiags, eye as speye
from scipy.sparse.linalg import spsolve
from menpofit.math.fft_utils import pad, crop


# TODO: Document me!
def mosse(X, y, l=0.01, boundary='constant', crop_filter=True):
    r"""
    Minimum Output Sum of Squared Errors (MOSSE) filter.

    Parameters
    ----------
    X : ``(n_images, n_channels, height, width)`` `ndarray`
        Training images.
    y : ``(1, height, width)`` `ndarray`
        Desired response.
    l: `float`, optional
        Regularization parameter.
    boundary: str {`constant`, `symmetric`}, optional
        Determines how the image is padded.
    crop_filter: `bool`, optional
        If ``True``, the shape of the MOSSE filter is the same as the shape
        of the desired response. If ``False``, the filter's shape is equal to:
        ``X[0].shape + y.shape - 1``

    Returns
    -------
    mosse: ``(1, height, width)`` `ndarray`
        Minimum Output Sum od Squared Errors (MOSSE) filter associated to
        the training images.

    References
    ----------
    .. [1] David S. Bolme, J. Ross Beveridge,  Bruce A. Draper and Yui Man Lui.
    "Visual Object Tracking using Adaptive Correlation Filters". CVPR, 2010.
    """
    # number of images, number of channels, height and width
    n, k, hx, wx = X.shape

    # height and width of desired responses
    _, hy, wy = y.shape
    y_shape = (hy, wy)

    # extended shape
    ext_h = hx + hy - 1
    ext_w = wx + wy - 1
    ext_shape = (ext_h, ext_w)

    # extend desired response
    ext_y = pad(y, ext_shape)
    # fft of extended desired response
    fft_ext_y = fft2(ext_y)

    # auto and cross spectral energy matrices
    sXX = 0
    sXY = 0
    # for each training image and desired response
    for x in X:
        # extend image
        ext_x = pad(x, ext_shape, boundary=boundary)
        # fft of extended image
        fft_ext_x = fft2(ext_x)

        # update auto and cross spectral energy matrices
        sXX += fft_ext_x.conj() * fft_ext_x
        sXY += fft_ext_x.conj() * fft_ext_y

    # compute desired correlation filter
    fft_ext_f = sXY / (sXX + l)
    # reshape extended filter to extended image shape
    fft_ext_f = fft_ext_f.reshape((k, ext_h, ext_w))

    # compute extended filter inverse fft
    f = np.real(ifftshift(ifft2(fft_ext_f), axes=(-2, -1)))

    if crop_filter:
        # crop extended filter to match desired response shape
        f = crop(f, y_shape)

    return f, sXY, sXX


def imosse(A, B, n_ab, X, y, l=0.01, boundary='constant',
           crop_filter=True, f=1.0):
    r"""
    Incremental Minimum Output Sum od Squared Errors (iMOSSE) filter

    Parameters
    ----------
    A :
    B :
    n_ab : `int`
        Total number of samples used to produce A and B.
    X : ``(n_images, n_channels, height, width)`` `ndarray`
        Training images.
    y : ``(1, height, width)`` `ndarray`
        Desired response.
    l : `float`, optional
        Regularization parameter.
    boundary : str {`constant`, `symmetric`}, optional
        Determines how the image is padded.
    crop_filter : `bool`, optional
    f : ``[0, 1]`` `float`, optional
        Forgetting factor that weights the relative contribution of new
        samples vs old samples. If 1.0, all samples are weighted equally.
        If <1.0, more emphasis is put on the new samples.

    Returns
    -------
    mccf : ``(1, height, width)`` `ndarray`
        Multi-Channel Correlation Filter (MCCF) filter associated to the
        training images.
    sXY :
    sXX :

    References
    ----------
    .. [1] David S. Bolme, J. Ross Beveridge,  Bruce A. Draper and Yui Man Lui.
    "Visual Object Tracking using Adaptive Correlation Filters". CVPR, 2010.
    """
    # number of images; number of channels, height and width
    n_x, k, hz, wz = X.shape

    # height and width of desired responses
    _, hy, wy = y.shape
    y_shape = (hy, wy)

    # multiply the number of samples used to produce the auto and cross
    # spectral energy matrices A and B by forgetting factor
    n_ab *= f
    # total number of samples
    n = n_ab + n_x
    # compute weighting factors
    nu_ab = n_ab / n
    nu_x = n_x / n

    # extended shape
    ext_h = hz + hy - 1
    ext_w = wz + wy - 1
    ext_shape = (ext_h, ext_w)

    # extend desired response
    ext_y = pad(y, ext_shape)
    # fft of extended desired response
    fft_ext_y = fft2(ext_y)

    # extend images
    ext_X = pad(X, ext_shape, boundary=boundary)

    # auto and cross spectral energy matrices
    sXX = 0
    sXY = 0
    # for each training image and desired response
    for ext_x in ext_X:
        # fft of extended image
        fft_ext_x = fft2(ext_x)

        # update auto and cross spectral energy matrices
        sXX += fft_ext_x.conj() * fft_ext_x
        sXY += fft_ext_x.conj() * fft_ext_y

    # combine old and new auto and cross spectral energy matrices
    sXY = nu_ab * A + nu_x * sXY
    sXX = nu_ab * B + nu_x * sXX
    # compute desired correlation filter
    fft_ext_f = sXY / (sXX + l)
    # reshape extended filter to extended image shape
    fft_ext_f = fft_ext_f.reshape((k, ext_h, ext_w))

    # compute filter inverse fft
    f = np.real(ifftshift(ifft2(fft_ext_f), axes=(-2, -1)))

    if crop_filter:
        # crop extended filter to match desired response shape
        f = crop(f, y_shape)

    return f, sXY, sXX


# TODO: Document me!
def mccf(X, y, l=0.01, boundary='constant', crop_filter=True):
    r"""
    Multi-Channel Correlation Filter (MCCF).

    Parameters
    ----------
    X : ``(n_images, n_channels, height, width)`` `ndarray`
        Training images.
    y : ``(1, height, width)`` `ndarray`
        Desired response.
    l : `float`, optional
        Regularization parameter.
    boundary : str {`constant`, `symmetric`}, optional
        Determines how the image is padded.
    crop_filter : `bool`, optional

    Returns
    -------
    mccf: ``(1, height, width)`` `ndarray`
        Multi-Channel Correlation Filter (MCCF) filter associated to the
        training images.
    sXY :
    sXX :

    References
    ----------
    .. [1] Hamed Kiani Galoogahi, Terence Sim,  Simon Lucey. "Multi-Channel
    Correlation Filters". ICCV, 2013.
    """
    # number of images; number of channels, height and width
    n, k, hx, wx = X.shape

    # height and width of desired responses
    _, hy, wy = y.shape
    y_shape = (hy, wy)

    # extended shape
    ext_h = hx + hy - 1
    ext_w = wx + wy - 1
    ext_shape = (ext_h, ext_w)
    # extended dimensionality
    ext_d = ext_h * ext_w

    # extend desired response
    ext_y = pad(y, ext_shape)
    # fft of extended desired response
    fft_ext_y = fft2(ext_y)

    # extend images
    ext_X = pad(X, ext_shape, boundary=boundary)

    # auto and cross spectral energy matrices
    sXX = 0
    sXY = 0
    # for each training image and desired response
    for ext_x in ext_X:
        # fft of extended image
        fft_ext_x = fft2(ext_x)

        # store extended image fft as sparse diagonal matrix
        diag_fft_x = spdiags(fft_ext_x.reshape((k, -1)),
                             -np.arange(0, k) * ext_d, ext_d * k, ext_d).T
        # vectorize extended desired response fft
        diag_fft_y = fft_ext_y.ravel()

        # update auto and cross spectral energy matrices
        sXX += diag_fft_x.conj().T.dot(diag_fft_x)
        sXY += diag_fft_x.conj().T.dot(diag_fft_y)

    # solve ext_d independent k x k linear systems (with regularization)
    # to obtain desired extended multi-channel correlation filter
    fft_ext_f = spsolve(sXX + l * speye(sXX.shape[-1]), sXY)
    # reshape extended filter to extended image shape
    fft_ext_f = fft_ext_f.reshape((k, ext_h, ext_w))

    # compute filter inverse fft
    f = np.real(ifftshift(ifft2(fft_ext_f), axes=(-2, -1)))

    if crop_filter:
        # crop extended filter to match desired response shape
        f = crop(f, y_shape)

    return f, sXY, sXX


# TODO: Document me!
def imccf(A, B, n_ab, X, y, l=0.01, boundary='constant', crop_filter=True,
          f=1.0):
    r"""
    Incremental Multi-Channel Correlation Filter (MCCF)

    Parameters
    ----------
    A :
    B :
    n_ab : `int`
        Total number of samples used to produce A and B.
    X : ``(n_images, n_channels, height, width)`` `ndarray`
        Training images.
    y : ``(1, height, width)`` `ndarray`
        Desired response.
    l : `float`, optional
        Regularization parameter.
    boundary : str {`constant`, `symmetric`}, optional
        Determines how the image is padded.
    crop_filter : `bool`, optional
    f : ``[0, 1]`` `float`, optional
        Forgetting factor that weights the relative contribution of new
        samples vs old samples. If 1.0, all samples are weighted equally.
        If <1.0, more emphasis is put on the new samples.

    Returns
    -------
    mccf : ``(1, height, width)`` `ndarray`
        Multi-Channel Correlation Filter (MCCF) filter associated to the
        training images.
    sXY :
    sXX :

    References
    ----------
    .. [1] David S. Bolme, J. Ross Beveridge,  Bruce A. Draper and Yui Man Lui.
    "Visual Object Tracking using Adaptive Correlation Filters". CVPR, 2010.
    .. [2] Hamed Kiani Galoogahi, Terence Sim,  Simon Lucey. "Multi-Channel
    Correlation Filters". ICCV, 2013.
    """
    # number of images; number of channels, height and width
    n_x, k, hz, wz = X.shape

    # height and width of desired responses
    _, hy, wy = y.shape
    y_shape = (hy, wy)

    # multiply the number of samples used to produce the auto and cross
    # spectral energy matrices A and B by forgetting factor
    n_ab *= f
    # total number of samples
    n = n_ab + n_x
    # compute weighting factors
    nu_ab = n_ab / n
    nu_x = n_x / n

    # extended shape
    ext_h = hz + hy - 1
    ext_w = wz + wy - 1
    ext_shape = (ext_h, ext_w)
    # extended dimensionality
    ext_d = ext_h * ext_w

    # extend desired response
    ext_y = pad(y, ext_shape)
    # fft of extended desired response
    fft_ext_y = fft2(ext_y)

    # extend images
    ext_X = pad(X, ext_shape, boundary=boundary)

    # auto and cross spectral energy matrices
    sXX = 0
    sXY = 0
    # for each training image and desired response
    for ext_x in ext_X:
        # fft of extended image
        fft_ext_x = fft2(ext_x)

        # store extended image fft as sparse diagonal matrix
        diag_fft_x = spdiags(fft_ext_x.reshape((k, -1)),
                             -np.arange(0, k) * ext_d, ext_d * k, ext_d).T
        # vectorize extended desired response fft
        diag_fft_y = fft_ext_y.ravel()

        # update auto and cross spectral energy matrices
        sXX += diag_fft_x.conj().T.dot(diag_fft_x)
        sXY += diag_fft_x.conj().T.dot(diag_fft_y)

    # combine old and new auto and cross spectral energy matrices
    sXY = nu_ab * A + nu_x * sXY
    sXX = nu_ab * B + nu_x * sXX
    # solve ext_d independent k x k linear systems (with regularization)
    # to obtain desired extended multi-channel correlation filter
    fft_ext_f = spsolve(sXX + l * speye(sXX.shape[-1]), sXY)
    # reshape extended filter to extended image shape
    fft_ext_f = fft_ext_f.reshape((k, ext_h, ext_w))

    # compute filter inverse fft
    f = np.real(ifftshift(ifft2(fft_ext_f), axes=(-2, -1)))
    if crop_filter:
        # crop extended filter to match desired response shape
        f = crop(f, y_shape)

    return f, sXY, sXX



from __future__ import division
from menpo.feature import no_op
from menpofit.transform import DifferentiableAlignmentAffine
from menpofit.fitter import (MultiFitter, noisy_shape_from_shape,
                             noisy_shape_from_bounding_box)
from menpofit import checks
from .algorithm import InverseCompositional
from .residual import SSD
from .result import LucasKanadeFitterResult


# TODO: document me!
class LucasKanadeFitter(MultiFitter):
    r"""
    """
    def __init__(self, template, group=None, holistic_features=no_op,
                 transform_cls=DifferentiableAlignmentAffine, diagonal=None,
                 scales=(0.5, 1.0), algorithm_cls=InverseCompositional,
                 residual_cls=SSD):

        checks.check_diagonal(diagonal)
        scales = checks.check_scales(scales)
        holistic_features = checks.check_callable(holistic_features,
                                                  len(scales))

        self.holistic_features = holistic_features
        self.transform_cls = transform_cls
        self.diagonal = diagonal
        self.scales = list(scales)
        # Make template masked for warping
        template = template.as_masked(copy=False)

        if self.diagonal:
            template = template.rescale_landmarks_to_diagonal_range(
                self.diagonal, group=group)
        self.reference_shape = template.landmarks[group].lms

        self.templates, self.sources = self._prepare_template(template,
                                                              group=group)
        self._set_up(algorithm_cls, residual_cls)

    def _set_up(self, algorithm_cls, residual_cls):
        self.algorithms = []
        for j, (t, s) in enumerate(zip(self.templates, self.sources)):
            transform = self.transform_cls(s, s)
            residual = residual_cls()
            algorithm = algorithm_cls(t, transform, residual)
            self.algorithms.append(algorithm)

    def _prepare_template(self, template, group=None):
        gt_shape = template.landmarks[group].lms
        templates, _, sources = self._prepare_image(template, gt_shape,
                                                    gt_shape=gt_shape)
        return templates, sources

    def perturb_from_gt_bb(self, gt_bb,
                           perturb_func=noisy_shape_from_bounding_box):
        return perturb_func(gt_bb, gt_bb)

    def _fitter_result(self, image, algorithm_results, affine_correction,
                       gt_shape=None):
        return LucasKanadeFitterResult(image, self, algorithm_results,
                                       affine_correction, gt_shape=gt_shape)



from functools import partial
from menpo.feature import no_op
from menpofit.result import euclidean_bb_normalised_error

from .base import (BaseSupervisedDescentAlgorithm,
                   compute_non_parametric_delta_x, features_per_image,
                   features_per_patch, update_non_parametric_estimates,
                   print_non_parametric_info, fit_non_parametric_shape)
from menpofit.math import IIRLRegression, IRLRegression, PCRRegression, \
    OptimalLinearRegression, OPPRegression


class NonParametricSDAlgorithm(BaseSupervisedDescentAlgorithm):
    r"""
    """

    def __init__(self):
        super(NonParametricSDAlgorithm, self).__init__()
        self.regressors = []

    def _compute_delta_x(self, gt_shapes, current_shapes):
        return compute_non_parametric_delta_x(gt_shapes, current_shapes)

    def _update_estimates(self, estimated_delta_x, delta_x, gt_x,
                          current_shapes):
        update_non_parametric_estimates(estimated_delta_x, delta_x, gt_x,
                                        current_shapes)

    def _compute_training_features(self, images, gt_shapes, current_shapes,
                                   prefix='', verbose=False):
        return features_per_image(images, current_shapes, self.patch_shape,
                                  self.patch_features, prefix=prefix,
                                  verbose=verbose)

    def _compute_test_features(self, image, current_shape):
        return features_per_patch(image, current_shape,
                                  self.patch_shape, self.patch_features)

    def run(self, image, initial_shape, gt_shape=None, **kwargs):
        return fit_non_parametric_shape(image, initial_shape, self,
                                        gt_shape=gt_shape)

    def _print_regression_info(self, template_shape, gt_shapes, n_perturbations,
                               delta_x, estimated_delta_x, level_index,
                               prefix=''):
        print_non_parametric_info(template_shape, gt_shapes, n_perturbations,
                                  delta_x, estimated_delta_x, level_index,
                                  self._compute_error, prefix=prefix)


# TODO: document me!
class NonParametricNewton(NonParametricSDAlgorithm):
    r"""
    """

    def __init__(self, patch_features=no_op, patch_shape=(17, 17),
                 n_iterations=3,
                 compute_error=euclidean_bb_normalised_error,
                 eps=10 ** -5, alpha=0, bias=True):
        super(NonParametricNewton, self).__init__()

        self._regressor_cls = partial(IRLRegression, alpha=alpha, bias=bias)
        self.patch_shape = patch_shape
        self.patch_features = patch_features
        self.n_iterations = n_iterations
        self._compute_error = compute_error
        self.eps = eps


# TODO: document me!
class NonParametricGaussNewton(NonParametricSDAlgorithm):
    r"""
    """

    def __init__(self, patch_features=no_op, patch_shape=(17, 17),
                 n_iterations=3,
                 compute_error=euclidean_bb_normalised_error,
                 eps=10 ** -5, alpha=0, bias=True, alpha2=0):
        super(NonParametricGaussNewton, self).__init__()

        self._regressor_cls = partial(IIRLRegression, alpha=alpha, bias=bias,
                                      alpha2=alpha2)
        self.patch_shape = patch_shape
        self.patch_features = patch_features
        self.n_iterations = n_iterations
        self._compute_error = compute_error
        self.eps = eps


class NonParametricPCRRegression(NonParametricSDAlgorithm):
    r"""
    """

    def __init__(self, patch_features=no_op, patch_shape=(17, 17),
                 n_iterations=3,
                 compute_error=euclidean_bb_normalised_error,
                 eps=10 ** -5, variance=None, bias=True):
        super(NonParametricPCRRegression, self).__init__()

        self._regressor_cls = partial(PCRRegression, variance=variance,
                                      bias=bias)
        self.patch_shape = patch_shape
        self.patch_features = patch_features
        self.n_iterations = n_iterations
        self._compute_error = compute_error
        self.eps = eps


class NonParametricOptimalRegression(NonParametricSDAlgorithm):
    r"""
    """

    def __init__(self, patch_features=no_op, patch_shape=(17, 17),
                 n_iterations=3,
                 compute_error=euclidean_bb_normalised_error,
                 eps=10 ** -5, variance=None, bias=True):
        super(NonParametricOptimalRegression, self).__init__()

        self._regressor_cls = partial(OptimalLinearRegression,
                                      variance=variance, bias=bias)
        self.patch_shape = patch_shape
        self.patch_features = patch_features
        self.n_iterations = n_iterations
        self._compute_error = compute_error
        self.eps = eps


class NonParametricOPPRegression(NonParametricSDAlgorithm):
    r"""
    """

    def __init__(self, patch_features=no_op, patch_shape=(17, 17),
                 n_iterations=3,
                 compute_error=euclidean_bb_normalised_error,
                 eps=10 ** -5, bias=True):
        super(NonParametricOPPRegression, self).__init__()

        self._regressor_cls = partial(OPPRegression, bias=bias)
        self.patch_shape = patch_shape
        self.patch_features = patch_features
        self.n_iterations = n_iterations
        self._compute_error = compute_error
        self.eps = eps



from scipy.linalg import norm
import numpy as np
from .result import LucasKanadeAlgorithmResult


# TODO: implement Inverse Additive Algorithm?
# TODO: implement sampling?
# TODO: document me!
class LucasKanade(object):
    r"""
    """
    def __init__(self, template, transform, residual, eps=10**-10):
        self.template = template
        self.transform = transform
        self.residual = residual
        self.eps = eps


# TODO: document me!
class ForwardAdditive(LucasKanade):
    r"""
    Forward Additive Lucas-Kanade algorithm
    """
    def run(self, image, initial_shape, max_iters=20, gt_shape=None):
        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        cost_functions = []

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Forward Compositional Algorithm
        while k < max_iters and eps > self.eps:
            # warp image
            IWxp = image.warp_to_mask(self.template.mask, self.transform,
                                      warp_landmarks=False)

            # compute warp jacobian
            dW_dp = np.rollaxis(
                self.transform.d_dp(self.template.indices()), -1)
            dW_dp = dW_dp.reshape(dW_dp.shape[:1] + self.template.shape +
                                  dW_dp.shape[-1:])

            # compute steepest descent images
            filtered_J, J = self.residual.steepest_descent_images(
                image, dW_dp, forward=(self.template, self.transform))

            # compute hessian
            H = self.residual.hessian(filtered_J, sdi2=J)

            # compute steepest descent parameter updates.
            sd_dp = self.residual.steepest_descent_update(
                filtered_J, IWxp, self.template)

            # compute gradient descent parameter updates
            dp = -np.real(np.linalg.solve(H, sd_dp))

            # Update warp weights
            self.transform.from_vector_inplace(self.transform.as_vector() + dp)
            p_list.append(self.transform.as_vector())

            # update cost
            cost_functions.append(self.residual.cost_closure())

            # test convergence
            eps = np.abs(norm(dp))

            # increase iteration counter
            k += 1

        return LucasKanadeAlgorithmResult(image, self, p_list,
                                          cost_functions=cost_functions,
                                          gt_shape=gt_shape)


# TODO: document me!
class ForwardCompositional(LucasKanade):
    r"""
    Forward Compositional Lucas-Kanade algorithm
    """
    def __init__(self, template, transform, residual, eps=10**-10):
        super(ForwardCompositional, self).__init__(
            template, transform, residual, eps=eps)
        self._precompute()

    def _precompute(self):
        # compute warp jacobian
        dW_dp = np.rollaxis(
            self.transform.d_dp(self.template.indices()), -1)
        self.dW_dp = dW_dp.reshape(dW_dp.shape[:1] + self.template.shape +
                                   dW_dp.shape[-1:])

    def run(self, image, initial_shape, max_iters=20, gt_shape=None):
        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        cost_functions = []

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Forward Compositional Algorithm
        while k < max_iters and eps > self.eps:
            # warp image
            IWxp = image.warp_to_mask(self.template.mask, self.transform,
                                      warp_landmarks=False)

            # compute steepest descent images
            filtered_J, J = self.residual.steepest_descent_images(
                IWxp, self.dW_dp)

            # compute hessian
            H = self.residual.hessian(filtered_J, sdi2=J)

            # compute steepest descent parameter updates.
            sd_dp = self.residual.steepest_descent_update(
                filtered_J, IWxp, self.template)

            # compute gradient descent parameter updates
            dp = -np.real(np.linalg.solve(H, sd_dp))

            # Update warp weights
            self.transform.compose_after_from_vector_inplace(dp)
            p_list.append(self.transform.as_vector())

            # update cost
            cost_functions.append(self.residual.cost_closure())

            # test convergence
            eps = np.abs(norm(dp))

            # increase iteration counter
            k += 1

        return LucasKanadeAlgorithmResult(image, self, p_list,
                                          cost_functions=cost_functions,
                                          gt_shape=gt_shape)


# TODO: document me!
class InverseCompositional(LucasKanade):
    r"""
    Inverse Compositional Lucas-Kanade algorithm
    """
    def __init__(self, template, transform, residual, eps=10**-10):
        super(InverseCompositional, self).__init__(
            template, transform, residual, eps=eps)
        self._precompute()

    def _precompute(self):
        # compute warp jacobian
        dW_dp = np.rollaxis(self.transform.d_dp(self.template.indices()), -1)
        dW_dp = dW_dp.reshape(dW_dp.shape[:1] + self.template.shape +
                              dW_dp.shape[-1:])
        # compute steepest descent images
        self.filtered_J, J = self.residual.steepest_descent_images(
            self.template, dW_dp)
        # compute hessian
        self.H = self.residual.hessian(self.filtered_J, sdi2=J)

    def run(self, image, initial_shape, max_iters=20, gt_shape=None):
        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        cost_functions = []

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Baker-Matthews, Inverse Compositional Algorithm
        while k < max_iters and eps > self.eps:
            # warp image
            IWxp = image.warp_to_mask(self.template.mask, self.transform,
                                      warp_landmarks=False)

            # compute steepest descent parameter updates.
            sd_dp = self.residual.steepest_descent_update(
                self.filtered_J, IWxp, self.template)

            # compute gradient descent parameter updates
            dp = np.real(np.linalg.solve(self.H, sd_dp))

            # update warp
            inv_dp = self.transform.pseudoinverse_vector(dp)
            self.transform.compose_after_from_vector_inplace(inv_dp)
            p_list.append(self.transform.as_vector())

            # update cost
            cost_functions.append(self.residual.cost_closure())

            # test convergence
            eps = np.abs(norm(dp))

            # increase iteration counter
            k += 1

        return LucasKanadeAlgorithmResult(image, self, p_list,
                                          cost_functions=cost_functions,
                                          gt_shape=gt_shape)



from __future__ import division
from menpofit.result import ParametricAlgorithmResult, MultiFitterResult


# TODO: document me!
class ATMAlgorithmResult(ParametricAlgorithmResult):
    r"""
    """
    def __init__(self, image, algorithm, shape_parameters, cost_functions=None,
                 gt_shape=None):
        super(ATMAlgorithmResult, self).__init__(
            image, algorithm, shape_parameters, gt_shape=gt_shape)
        self._cost_functions = cost_functions
        self._warped_images = None
        self._costs = None

    @property
    def warped_images(self):
        if self._warped_images is None:
            self._warped_images = []
            for p in self.shape_parameters:
                self.algorithm.transform._from_vector_inplace(p)
                self._warped_images.append(
                    self.algorithm.warp(self.image))
        return self._warped_images

    @property
    def costs(self):
        if self._costs is None:
            self._costs = [f() for f in self._cost_functions]
        return self._costs


# TODO: document me!
class LinearATMAlgorithmResult(ATMAlgorithmResult):
    r"""
    """
    @property
    def shapes(self):
        return [self.algorithm.transform.from_vector(p).sparse_target
                for p in self.shape_parameters]

    @property
    def final_shape(self):
        return self.final_transform.sparse_target

    @property
    def initial_shape(self):
        return self.initial_transform.sparse_target


# TODO: document me!
class ATMFitterResult(MultiFitterResult):
    r"""
    """
    def __init__(self, image, fitter, algorithm_results, affine_correction,
                 gt_shape=None):
        super(ATMFitterResult, self).__init__(
            image, fitter, algorithm_results, affine_correction,
            gt_shape=gt_shape)
        self._warped_images = None

    @property
    def warped_images(self):
        if self._warped_images is None:
            algorithm = self.algorithm_results[-1].algorithm
            self._warped_images = []
            for s in self.shapes:
                algorithm.transform.set_target(s)
                self._warped_images.append(
                    algorithm.warp(self.image))
        return self._warped_images

    @property
    def costs(self):
        costs = []
        for a in self.algorithm_results:
            costs += a.costs
        return costs



from __future__ import division
import numpy as np
from menpo.image import Image
from menpo.feature import gradient as fast_gradient, no_op
from ..result import AAMAlgorithmResult, LinearAAMAlgorithmResult


def _solve_all_map(H, J, e, Ja_prior, c, Js_prior, p, m, n):
    if n is not H.shape[0] - m:
        # Bidirectional Compositional case
        Js_prior = np.hstack((Js_prior, Js_prior))
        p = np.hstack((p, p))
        # compute and return MAP solution
    J_prior = np.hstack((Ja_prior, Js_prior))
    H += np.diag(J_prior)
    Je = J_prior * np.hstack((c, p)) + J.T.dot(e)
    dq = - np.linalg.solve(H, Je)
    return dq[:m], dq[m:]


def _solve_all_ml(H, J, e, m):
    # compute ML solution
    dq = - np.linalg.solve(H, J.T.dot(e))
    return dq[:m], dq[m:]


# TODO document me!
class LucasKanadeBaseInterface(object):
    r"""
    """
    def __init__(self, transform, template, sampling=None):
        self.transform = transform
        self.template = template

        self._build_sampling_mask(sampling)

    def _build_sampling_mask(self, sampling):
        n_true_pixels = self.template.n_true_pixels()
        n_channels = self.template.n_channels
        n_parameters = self.transform.n_parameters

        sampling_mask = np.zeros(n_true_pixels, dtype=np.bool)

        if sampling is None:
            sampling = range(0, n_true_pixels, 1)
        elif isinstance(sampling, np.int):
            sampling = range(0, n_true_pixels, sampling)

        sampling_mask[sampling] = 1

        self.i_mask = np.nonzero(np.tile(
            sampling_mask[None, ...], (n_channels, 1)).flatten())[0]
        self.dW_dp_mask = np.nonzero(np.tile(
            sampling_mask[None, ..., None], (2, 1, n_parameters)))
        self.nabla_mask = np.nonzero(np.tile(
            sampling_mask[None, None, ...], (2, n_channels, 1)))
        self.nabla2_mask = np.nonzero(np.tile(
            sampling_mask[None, None, None, ...], (2, 2, n_channels, 1)))

    @property
    def shape_model(self):
        return self.transform.pdm.model

    @property
    def n(self):
        return self.transform.n_parameters

    @property
    def true_indices(self):
        return self.template.mask.true_indices()

    def warp_jacobian(self):
        dW_dp = np.rollaxis(self.transform.d_dp(self.true_indices), -1)
        return dW_dp[self.dW_dp_mask].reshape((dW_dp.shape[0], -1,
                                               dW_dp.shape[2]))

    def warp(self, image):
        return image.warp_to_mask(self.template.mask, self.transform,
                                  warp_landmarks=False)

    def gradient(self, img):
        nabla = fast_gradient(img)
        nabla.set_boundary_pixels()
        return nabla.as_vector().reshape((2, img.n_channels, -1))

    def steepest_descent_images(self, nabla, dW_dp):
        # reshape gradient
        # nabla: n_dims x n_channels x n_pixels
        nabla = nabla[self.nabla_mask].reshape(nabla.shape[:2] + (-1,))
        # compute steepest descent images
        # nabla: n_dims x n_channels x n_pixels
        # warp_jacobian: n_dims x            x n_pixels x n_params
        # sdi:            n_channels x n_pixels x n_params
        sdi = 0
        a = nabla[..., None] * dW_dp[:, None, ...]
        for d in a:
            sdi += d
        # reshape steepest descent images
        # sdi: (n_channels x n_pixels) x n_params
        return sdi.reshape((-1, sdi.shape[2]))

    @classmethod
    def solve_shape_map(cls, H, J, e, J_prior, p):
        if p.shape[0] is not H.shape[0]:
            # Bidirectional Compositional case
            J_prior = np.hstack((J_prior, J_prior))
            p = np.hstack((p, p))
        # compute and return MAP solution
        H += np.diag(J_prior)
        Je = J_prior * p + J.T.dot(e)
        return - np.linalg.solve(H, Je)

    @classmethod
    def solve_shape_ml(cls, H, J, e):
        # compute and return ML solution
        return -np.linalg.solve(H, J.T.dot(e))


class LucasKanadeStandardInterface(LucasKanadeBaseInterface):

    def __init__(self, appearance_model, transform, template, sampling=None):
        super(LucasKanadeStandardInterface, self).__init__(transform, template,
                                                           sampling=sampling)
        self.appearance_model = appearance_model

    @property
    def m(self):
        return self.appearance_model.n_active_components

    def solve_all_map(self, H, J, e, Ja_prior, c, Js_prior, p):
        return _solve_all_map(H, J, e, Ja_prior, c, Js_prior, p,
                              self.m, self.n)

    def solve_all_ml(self, H, J, e):
        return _solve_all_ml(H, J, e, self.m)

    def algorithm_result(self, image, shape_parameters, cost_functions=None,
                         appearance_parameters=None, gt_shape=None):
        return AAMAlgorithmResult(
            image, self, shape_parameters,
            cost_functions=cost_functions,
            appearance_parameters=appearance_parameters, gt_shape=gt_shape)


# TODO document me!
class LucasKanadeLinearInterface(LucasKanadeStandardInterface):
    r"""
    """
    @property
    def shape_model(self):
        return self.transform.model

    def algorithm_result(self, image, shape_parameters, cost_functions=None,
                         appearance_parameters=None, gt_shape=None):
        return LinearAAMAlgorithmResult(
            image, self, shape_parameters,
            cost_functions=cost_functions,
            appearance_parameters=appearance_parameters, gt_shape=gt_shape)


# TODO document me!
class LucasKanadePatchBaseInterface(LucasKanadeBaseInterface):
    r"""
    """
    def __init__(self, transform, template, sampling=None,
                 patch_shape=(17, 17), patch_normalisation=no_op):
        self.patch_shape = patch_shape
        self.patch_normalisation = patch_normalisation

        super(LucasKanadePatchBaseInterface, self).__init__(
            transform, template, sampling=sampling)

    def _build_sampling_mask(self, sampling):
        if sampling is None:
            sampling = np.ones(self.patch_shape, dtype=np.bool)

        image_shape = self.template.pixels.shape
        image_mask = np.tile(sampling[None, None, None, ...],
                             image_shape[:3] + (1, 1))
        self.i_mask = np.nonzero(image_mask.flatten())[0]
        self.gradient_mask = np.nonzero(np.tile(
            image_mask[None, ...], (2, 1, 1, 1, 1, 1)))
        self.gradient2_mask = np.nonzero(np.tile(
            image_mask[None, None, ...], (2, 2, 1, 1, 1, 1, 1)))

    @property
    def shape_model(self):
        return self.transform.model

    def warp_jacobian(self):
        return np.rollaxis(self.transform.d_dp(None), -1)

    def warp(self, image):
        parts = image.extract_patches(self.transform.target,
                                      patch_shape=self.patch_shape,
                                      as_single_array=True)
        parts = self.patch_normalisation(parts)
        return Image(parts, copy=False)

    def gradient(self, image):
        pixels = image.pixels
        nabla = fast_gradient(pixels.reshape((-1,) + self.patch_shape))
        # remove 1st dimension gradient which corresponds to the gradient
        # between parts
        return nabla.reshape((2,) + pixels.shape)

    def steepest_descent_images(self, nabla, dw_dp):
        # reshape nabla
        # nabla: dims x parts x off x ch x (h x w)
        nabla = nabla[self.gradient_mask].reshape(
            nabla.shape[:-2] + (-1,))
        # compute steepest descent images
        # nabla: dims x parts x off x ch x (h x w)
        # ds_dp:    dims x parts x                             x params
        # sdi:             parts x off x ch x (h x w) x params
        sdi = 0
        a = nabla[..., None] * dw_dp[..., None, None, None, :]
        for d in a:
            sdi += d

        # reshape steepest descent images
        # sdi: (parts x offsets x ch x w x h) x params
        return sdi.reshape((-1, sdi.shape[-1]))


# TODO document me!
class LucasKanadePatchInterface(LucasKanadePatchBaseInterface):
    r"""
    """
    def __init__(self, appearance_model, transform, template, sampling=None,
                 patch_shape=(17, 17), patch_normalisation=no_op):
        self.appearance_model = appearance_model

        super(LucasKanadePatchInterface, self).__init__(
            transform, template, patch_shape=patch_shape,
            patch_normalisation=patch_normalisation, sampling=sampling)

    @property
    def m(self):
        return self.appearance_model.n_active_components

    def solve_all_map(self, H, J, e, Ja_prior, c, Js_prior, p):
        return _solve_all_map(H, J, e, Ja_prior, c, Js_prior, p,
                              self.m, self.n)

    def solve_all_ml(self, H, J, e):
        return _solve_all_ml(H, J, e, self.m)

    def algorithm_result(self, image, shape_parameters, cost_functions=None,
                         appearance_parameters=None, gt_shape=None):
        return AAMAlgorithmResult(
            image, self, shape_parameters,
            cost_functions=cost_functions,
            appearance_parameters=appearance_parameters, gt_shape=gt_shape)


# TODO document me!
class LucasKanade(object):
    r"""
    """
    def __init__(self, aam_interface, eps=10**-5):
        self.eps = eps
        self.interface = aam_interface
        self._precompute()

    @property
    def appearance_model(self):
        return self.interface.appearance_model

    @property
    def transform(self):
        return self.interface.transform

    @property
    def template(self):
        return self.interface.template

    def _precompute(self):
        # grab number of shape and appearance parameters
        self.n = self.transform.n_parameters
        self.m = self.appearance_model.n_active_components

        # grab appearance model components
        self.A = self.appearance_model.components
        # mask them
        self.A_m = self.A.T[self.interface.i_mask, :]
        # compute their pseudoinverse
        self.pinv_A_m = np.linalg.pinv(self.A_m)

        # grab appearance model mean
        self.a_bar = self.appearance_model.mean()
        # vectorize it and mask it
        self.a_bar_m = self.a_bar.as_vector()[self.interface.i_mask]

        # compute warp jacobian
        self.dW_dp = self.interface.warp_jacobian()

        # compute shape model prior
        # TODO: Is this correct? It's like modelling no noise at all
        sm_noise_variance = self.interface.shape_model.noise_variance() or 1
        s2 = self.appearance_model.noise_variance() / sm_noise_variance
        L = self.interface.shape_model.eigenvalues
        self.s2_inv_L = np.hstack((np.ones((4,)), s2 / L))
        # compute appearance model prior
        S = self.appearance_model.eigenvalues
        self.s2_inv_S = s2 / S


# TODO: Document me!
class ProjectOut(LucasKanade):
    r"""
    Abstract Interface for Project-out AAM algorithms
    """
    def project_out(self, J):
        # project-out appearance bases from a particular vector or matrix
        return J - self.A_m.dot(self.pinv_A_m.dot(J))

    def run(self, image, initial_shape, gt_shape=None, max_iters=20,
            map_inference=False):
        # define cost closure
        def cost_closure(x, f):
            return lambda: x.T.dot(f(x))

        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Compositional Gauss-Newton loop -------------------------------------

        # warp image
        self.i = self.interface.warp(image)
        # vectorize it and mask it
        i_m = self.i.as_vector()[self.interface.i_mask]

        # compute masked error
        self.e_m = i_m - self.a_bar_m

        # update cost_functions
        cost_functions = [cost_closure(self.e_m, self.project_out)]

        while k < max_iters and eps > self.eps:
            # solve for increments on the shape parameters
            self.dp = self._solve(map_inference)

            # update warp
            s_k = self.transform.target.points
            self._update_warp()
            p_list.append(self.transform.as_vector())

            # warp image
            self.i = self.interface.warp(image)
            # vectorize it and mask it
            i_m = self.i.as_vector()[self.interface.i_mask]

            # compute masked error
            self.e_m = i_m - self.a_bar_m

            # update cost
            cost_functions.append(cost_closure(self.e_m, self.project_out))

            # test convergence
            eps = np.abs(np.linalg.norm(s_k - self.transform.target.points))

            # increase iteration counter
            k += 1

        # return algorithm result
        return self.interface.algorithm_result(
            image, p_list, cost_functions=cost_functions, gt_shape=gt_shape)


# TODO: Document me!
class ProjectOutForwardCompositional(ProjectOut):
    r"""
    Project-out Forward Compositional (PFC) Gauss-Newton algorithm
    """
    def _solve(self, map_inference):
        # compute warped image gradient
        nabla_i = self.interface.gradient(self.i)
        # compute masked forward Jacobian
        J_m = self.interface.steepest_descent_images(nabla_i, self.dW_dp)
        # project out appearance model from it
        QJ_m = self.project_out(J_m)
        # compute masked forward Hessian
        JQJ_m = QJ_m.T.dot(J_m)
        # solve for increments on the shape parameters
        if map_inference:
            return self.interface.solve_shape_map(
                JQJ_m, QJ_m, self.e_m,  self.s2_inv_L,
                self.transform.as_vector())
        else:
            return self.interface.solve_shape_ml(JQJ_m, QJ_m, self.e_m)

    def _update_warp(self):
        # update warp based on forward composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() + self.dp)


# TODO: Document me!
class ProjectOutInverseCompositional(ProjectOut):
    r"""
    Project-out Inverse Compositional (PIC) Gauss-Newton algorithm
    """
    def _precompute(self):
        # call super method
        super(ProjectOutInverseCompositional, self)._precompute()
        # compute appearance model mean gradient
        nabla_a = self.interface.gradient(self.a_bar)
        # compute masked inverse Jacobian
        J_m = self.interface.steepest_descent_images(-nabla_a, self.dW_dp)
        # project out appearance model from it
        self.QJ_m = self.project_out(J_m)
        # compute masked inverse Hessian
        self.JQJ_m = self.QJ_m.T.dot(J_m)
        # compute masked Jacobian pseudo-inverse
        self.pinv_QJ_m = np.linalg.solve(self.JQJ_m, self.QJ_m.T)

    def _solve(self, map_inference):
        # solve for increments on the shape parameters
        if map_inference:
            return self.interface.solve_shape_map(
                self.JQJ_m, self.QJ_m, self.e_m, self.s2_inv_L,
                self.transform.as_vector())
        else:
            return -self.pinv_QJ_m.dot(self.e_m)

    def _update_warp(self):
        # update warp based on inverse composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() - self.dp)


# TODO: Document me!
class Simultaneous(LucasKanade):
    r"""
    Abstract Interface for Simultaneous AAM algorithms
    """
    def run(self, image, initial_shape, gt_shape=None, max_iters=20,
            map_inference=False):
        # define cost closure
        def cost_closure(x):
            return lambda: x.T.dot(x)

        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Compositional Gauss-Newton loop -------------------------------------

        # warp image
        self.i = self.interface.warp(image)
        # mask warped image
        i_m = self.i.as_vector()[self.interface.i_mask]

        # initialize appearance parameters by projecting masked image
        # onto masked appearance model
        self.c = self.pinv_A_m.dot(i_m - self.a_bar_m)
        self.a = self.appearance_model.instance(self.c)
        a_m = self.a.as_vector()[self.interface.i_mask]
        c_list = [self.c]

        # compute masked error
        self.e_m = i_m - a_m

        # update cost
        cost_functions = [cost_closure(self.e_m)]

        while k < max_iters and eps > self.eps:
            # solve for increments on the appearance and shape parameters
            # simultaneously
            dc, self.dp = self._solve(map_inference)

            # update appearance parameters
            self.c = self.c + dc
            self.a = self.appearance_model.instance(self.c)
            a_m = self.a.as_vector()[self.interface.i_mask]
            c_list.append(self.c)

            # update warp
            s_k = self.transform.target.points
            self._update_warp()
            p_list.append(self.transform.as_vector())

            # warp image
            self.i = self.interface.warp(image)
            # mask warped image
            i_m = self.i.as_vector()[self.interface.i_mask]

            # compute masked error
            self.e_m = i_m - a_m

            # update cost
            cost_functions.append(cost_closure(self.e_m))

            # test convergence
            eps = np.abs(np.linalg.norm(s_k - self.transform.target.points))

            # increase iteration counter
            k += 1

        # return algorithm result
        return self.interface.algorithm_result(
            image, p_list, cost_functions=cost_functions,
            appearance_parameters=c_list, gt_shape=gt_shape)

    def _solve(self, map_inference):
        # compute masked Jacobian
        J_m = self._compute_jacobian()
        # assemble masked simultaneous Jacobian
        J_sim_m = np.hstack((-self.A_m, J_m))
        # compute masked Hessian
        H_sim_m = J_sim_m.T.dot(J_sim_m)
        # solve for increments on the appearance and shape parameters
        # simultaneously
        if map_inference:
            return self.interface.solve_all_map(
                H_sim_m, J_sim_m, self.e_m, self.s2_inv_S, self.c,
                self.s2_inv_L, self.transform.as_vector())
        else:
            return self.interface.solve_all_ml(H_sim_m, J_sim_m, self.e_m)


# TODO: Document me!
class SimultaneousForwardCompositional(Simultaneous):
    r"""
    Simultaneous Forward Compositional (SFC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped image gradient
        nabla_i = self.interface.gradient(self.i)
        # return forward Jacobian
        return self.interface.steepest_descent_images(nabla_i, self.dW_dp)

    def _update_warp(self):
        # update warp based on forward composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() + self.dp)


# TODO: Document me!
class SimultaneousInverseCompositional(Simultaneous):
    r"""
    Simultaneous Inverse Compositional (SIC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped appearance model gradient
        nabla_a = self.interface.gradient(self.a)
        # return inverse Jacobian
        return self.interface.steepest_descent_images(-nabla_a, self.dW_dp)

    def _update_warp(self):
        # update warp based on inverse composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() - self.dp)


# TODO: Document me!
class Alternating(LucasKanade):
    r"""
    Abstract Interface for Alternating AAM algorithms
    """
    def _precompute(self, **kwargs):
        # call super method
        super(Alternating, self)._precompute()
        # compute MAP appearance Hessian
        self.AA_m_map = self.A_m.T.dot(self.A_m) + np.diag(self.s2_inv_S)

    def run(self, image, initial_shape, gt_shape=None, max_iters=20,
            map_inference=False):
        # define cost closure
        def cost_closure(x):
            return lambda: x.T.dot(x)

        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Compositional Gauss-Newton loop -------------------------------------

        # warp image
        self.i = self.interface.warp(image)
        # mask warped image
        i_m = self.i.as_vector()[self.interface.i_mask]

        # initialize appearance parameters by projecting masked image
        # onto masked appearance model
        c = self.pinv_A_m.dot(i_m - self.a_bar_m)
        self.a = self.appearance_model.instance(c)
        a_m = self.a.as_vector()[self.interface.i_mask]
        c_list = [c]
        Jdp = 0

        # compute masked error
        e_m = i_m - a_m

        # update cost
        cost_functions = [cost_closure(e_m)]

        while k < max_iters and eps > self.eps:
            # solve for increment on the appearance parameters
            if map_inference:
                Ae_m_map = - self.s2_inv_S * c + self.A_m.dot(e_m + Jdp)
                dc = np.linalg.solve(self.AA_m_map, Ae_m_map)
            else:
                dc = self.pinv_A_m.dot(e_m + Jdp)

            # compute masked Jacobian
            J_m = self._compute_jacobian()
            # compute masked Hessian
            H_m = J_m.T.dot(J_m)
            # solve for increments on the shape parameters
            if map_inference:
                self.dp = self.interface.solve_shape_map(
                    H_m, J_m, e_m - self.A_m.T.dot(dc), self.s2_inv_L,
                    self.transform.as_vector())
            else:
                self.dp = self.interface.solve_shape_ml(H_m, J_m,
                                                        e_m - self.A_m.dot(dc))

            # update appearance parameters
            c = c + dc
            self.a = self.appearance_model.instance(c)
            a_m = self.a.as_vector()[self.interface.i_mask]
            c_list.append(c)

            # update warp
            s_k = self.transform.target.points
            self._update_warp()
            p_list.append(self.transform.as_vector())

            # warp image
            self.i = self.interface.warp(image)
            # mask warped image
            i_m = self.i.as_vector()[self.interface.i_mask]

            # compute Jdp
            Jdp = J_m.dot(self.dp)

            # compute masked error
            e_m = i_m - a_m

            # update cost
            cost_functions.append(cost_closure(e_m))

            # test convergence
            eps = np.abs(np.linalg.norm(s_k - self.transform.target.points))

            # increase iteration counter
            k += 1

        # return algorithm result
        return self.interface.algorithm_result(
            image, p_list, cost_functions=cost_functions,
            appearance_parameters=c_list, gt_shape=gt_shape)


# TODO: Document me!
class AlternatingForwardCompositional(Alternating):
    r"""
    Alternating Forward Compositional (AFC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped image gradient
        nabla_i = self.interface.gradient(self.i)
        # return forward Jacobian
        return self.interface.steepest_descent_images(nabla_i, self.dW_dp)

    def _update_warp(self):
        # update warp based on forward composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() + self.dp)


# TODO: Document me!
class AlternatingInverseCompositional(Alternating):
    r"""
    Alternating Inverse Compositional (AIC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped appearance model gradient
        nabla_a = self.interface.gradient(self.a)
        # return inverse Jacobian
        return self.interface.steepest_descent_images(-nabla_a, self.dW_dp)

    def _update_warp(self):
        # update warp based on inverse composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() - self.dp)


# TODO: Document me!
class ModifiedAlternating(Alternating):
    r"""
    Abstract Interface for Modified Alternating AAM algorithms
    """
    def run(self, image, initial_shape, gt_shape=None, max_iters=20,
            map_inference=False):
        # define cost closure
        def cost_closure(x):
            return lambda: x.T.dot(x)

        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        # initialize iteration counter and epsilon
        a_m = self.a_bar_m
        c_list = []
        k = 0
        eps = np.Inf

        # Compositional Gauss-Newton loop -------------------------------------

        # warp image
        self.i = self.interface.warp(image)
        # mask warped image
        i_m = self.i.as_vector()[self.interface.i_mask]

        # initialize appearance parameters by projecting masked image
        # onto masked appearance model
        c = self.pinv_A_m.dot(i_m - a_m)
        self.a = self.appearance_model.instance(c)
        a_m = self.a.as_vector()[self.interface.i_mask]
        c_list.append(c)

        # compute masked error
        e_m = i_m - a_m

        # update cost
        cost_functions = [cost_closure(e_m)]

        while k < max_iters and eps > self.eps:
            # compute masked Jacobian
            J_m = self._compute_jacobian()
            # compute masked Hessian
            H_m = J_m.T.dot(J_m)
            # solve for increments on the shape parameters
            if map_inference:
                self.dp = self.interface.solve_shape_map(
                    H_m, J_m, e_m, self.s2_inv_L, self.transform.as_vector())
            else:
                self.dp = self.interface.solve_shape_ml(H_m, J_m, e_m)

            # update warp
            s_k = self.transform.target.points
            self._update_warp()
            p_list.append(self.transform.as_vector())

            # warp image
            self.i = self.interface.warp(image)
            # mask warped image
            i_m = self.i.as_vector()[self.interface.i_mask]

            # update appearance parameters
            c = self.pinv_A_m.dot(i_m - self.a_bar_m)
            self.a = self.appearance_model.instance(c)
            a_m = self.a.as_vector()[self.interface.i_mask]
            c_list.append(c)

            # compute masked error
            e_m = i_m - a_m

            # update cost
            cost_functions.append(cost_closure(e_m))

            # test convergence
            eps = np.abs(np.linalg.norm(s_k - self.transform.target.points))

            # increase iteration counter
            k += 1

        # return algorithm result
        return self.interface.algorithm_result(
            image, p_list, cost_functions=cost_functions,
            appearance_parameters=c_list, gt_shape=gt_shape)


# TODO: Document me!
class ModifiedAlternatingForwardCompositional(ModifiedAlternating):
    r"""
    Modified Alternating Forward Compositional (MAFC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped image gradient
        nabla_i = self.interface.gradient(self.i)
        # return forward Jacobian
        return self.interface.steepest_descent_images(nabla_i, self.dW_dp)

    def _update_warp(self):
        # update warp based on forward composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() + self.dp)


# TODO: Document me!
class ModifiedAlternatingInverseCompositional(ModifiedAlternating):
    r"""
    Modified Alternating Inverse Compositional (MAIC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped appearance model gradient
        nabla_a = self.interface.gradient(self.a)
        # return inverse Jacobian
        return self.interface.steepest_descent_images(-nabla_a, self.dW_dp)

    def _update_warp(self):
        # update warp based on inverse composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() - self.dp)


# TODO: Document me!
class Wiberg(LucasKanade):
    r"""
    Abstract Interface for Wiberg AAM algorithms
    """
    def project_out(self, J):
        # project-out appearance bases from a particular vector or matrix
        return J - self.A_m.dot(self.pinv_A_m.dot(J))

    def run(self, image, initial_shape, gt_shape=None, max_iters=20,
            map_inference=False):
        # define cost closure
        def cost_closure(x, f):
            return lambda: x.T.dot(f(x))

        # initialize transform
        self.transform.set_target(initial_shape)
        p_list = [self.transform.as_vector()]

        # initialize iteration counter and epsilon
        k = 0
        eps = np.Inf

        # Compositional Gauss-Newton loop -------------------------------------

        # warp image
        self.i = self.interface.warp(image)
        # mask warped image
        i_m = self.i.as_vector()[self.interface.i_mask]

        # initialize appearance parameters by projecting masked image
        # onto masked appearance model
        c = self.pinv_A_m.dot(i_m - self.a_bar_m)
        self.a = self.appearance_model.instance(c)
        a_m = self.a.as_vector()[self.interface.i_mask]
        c_list = [c]

        # compute masked error
        e_m = i_m - self.a_bar_m

        # update cost
        cost_functions = [cost_closure(e_m, self.project_out)]

        while k < max_iters and eps > self.eps:
            # compute masked Jacobian
            J_m = self._compute_jacobian()
            # project out appearance models
            QJ_m = self.project_out(J_m)
            # compute masked Hessian
            JQJ_m = QJ_m.T.dot(J_m)
            # solve for increments on the shape parameters
            if map_inference:
                self.dp = self.interface.solve_shape_map(
                    JQJ_m, QJ_m, e_m, self.s2_inv_L,
                    self.transform.as_vector())
            else:
                self.dp = self.interface.solve_shape_ml(JQJ_m, QJ_m, e_m)

            # update warp
            s_k = self.transform.target.points
            self._update_warp()
            p_list.append(self.transform.as_vector())

            # warp image
            self.i = self.interface.warp(image)
            # mask warped image
            i_m = self.i.as_vector()[self.interface.i_mask]

            # update appearance parameters
            dc = self.pinv_A_m.dot(i_m - a_m + J_m.dot(self.dp))
            c = c + dc
            self.a = self.appearance_model.instance(c)
            a_m = self.a.as_vector()[self.interface.i_mask]
            c_list.append(c)

            # compute masked error
            e_m = i_m - self.a_bar_m

            # update cost
            cost_functions.append(cost_closure(e_m, self.project_out))

            # test convergence
            eps = np.abs(np.linalg.norm(s_k - self.transform.target.points))

            # increase iteration counter
            k += 1

        # return algorithm result
        return self.interface.algorithm_result(
            image, p_list, cost_functions=cost_functions,
            appearance_parameters=c_list, gt_shape=gt_shape)


# TODO: Document me!
class WibergForwardCompositional(Wiberg):
    r"""
    Wiberg Forward Compositional (WFC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped image gradient
        nabla_i = self.interface.gradient(self.i)
        # return forward Jacobian
        return self.interface.steepest_descent_images(nabla_i, self.dW_dp)

    def _update_warp(self):
        # update warp based on forward composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() + self.dp)


# TODO: Document me!
class WibergInverseCompositional(Wiberg):
    r"""
    Wiberg Inverse Compositional (WIC) Gauss-Newton algorithm
    """
    def _compute_jacobian(self):
        # compute warped appearance model gradient
        nabla_a = self.interface.gradient(self.a)
        # return inverse Jacobian
        return self.interface.steepest_descent_images(-nabla_a, self.dW_dp)

    def _update_warp(self):
        # update warp based on inverse composition
        self.transform._from_vector_inplace(
            self.transform.as_vector() - self.dp)



import numpy as np
from menpo.transform import (Affine, AlignmentAffine, Similarity,
                             AlignmentSimilarity)
from menpofit.differentiable import DP, DX


class DifferentiableAffine(Affine, DP, DX):

    def d_dp(self, points):
        return affine_d_dp(self, points)

    def d_dx(self, points):
        return affine_d_dx(self)


class DifferentiableAlignmentAffine(AlignmentAffine, DP, DX):

    def as_non_alignment(self):
        return DifferentiableAffine(self.h_matrix, skip_checks=True)

    def d_dp(self, points):
        return affine_d_dp(self, points)

    def d_dx(self, points):
        return affine_d_dx(self)


class DifferentiableSimilarity(Similarity, DP, DX):

    def d_dp(self, points):
        return similarity_d_dp(self, points)

    def d_dx(self, points):
        return affine_d_dx(self)


class DifferentiableAlignmentSimilarity(AlignmentSimilarity, DP, DX):

    def as_non_alignment(self):
        return DifferentiableSimilarity(self.h_matrix, skip_checks=True)

    def d_dp(self, points):
        return similarity_d_dp(self, points)

    def d_dx(self, points):
        return affine_d_dx(self)


def affine_d_dx(affine):
    r"""
    The first order derivative of an Affine transform wrt spatial changes
    evaluated at points.

    The Jacobian for a given point (for 2D) is of the form::

        Jx = [(1 + a),     -b  ]
        Jy = [   b,     (1 + a)]
        J =  [Jx, Jy] = [[(1 + a), -b], [b, (1 + a)]]

    where a and b come from:

        W(x;p) = [1 + a   -b      tx] [x]
                 [b       1 + a   ty] [y]
                                      [1]
    Hence it is simply the linear component of the transform.

    Returns
    -------

    d_dx: (1, n_dims, n_dims) ndarray
        The jacobian wrt spatial changes.

        d_dx[0, j, k] is the scalar differential change that the
        j'th dimension of the i'th point experiences due to a first order
        change in the k'th dimension.

        Note that because the jacobian is constant across space the first
        axis is length 1 to allow for broadcasting.

    """
    return affine.linear_component[None, ...]


def affine_d_dp(self, points):
    r"""The first order derivative of this Affine transform wrt parameter
    changes evaluated at points.

    The Jacobian generated (for 2D) is of the form::

        x 0 y 0 1 0
        0 x 0 y 0 1

    This maintains a parameter order of::

      W(x;p) = [1 + p1  p3      p5] [x]
               [p2      1 + p4  p6] [y]
                                    [1]

    Parameters
    ----------
    points : (n_points, n_dims) ndarray
        The set of points to calculate the jacobian for.


    Returns
    -------
    (n_points, n_params, n_dims) ndarray
        The jacobian wrt parametrization

    """
    n_points, points_n_dim = points.shape
    if points_n_dim != self.n_dims:
        raise ValueError(
            "Trying to sample jacobian in incorrect dimensions "
            "(transform is {0}D, sampling at {1}D)".format(
                self.n_dims, points_n_dim))
    # prealloc the jacobian
    jac = np.zeros((n_points, self.n_parameters, self.n_dims))
    # a mask that we can apply at each iteration
    dim_mask = np.eye(self.n_dims, dtype=np.bool)

    for i, s in enumerate(
            range(0, self.n_dims * self.n_dims, self.n_dims)):
        # i is current axis
        # s is slicing offset
        # make a mask for a single points jacobian
        full_mask = np.zeros((self.n_parameters, self.n_dims), dtype=bool)
        # fill the mask in for the ith axis
        full_mask[slice(s, s + self.n_dims)] = dim_mask
        # assign the ith axis points to this mask, broadcasting over all
        # points
        jac[:, full_mask] = points[:, i][..., None]
        # finally, just repeat the same but for the ones at the end
    full_mask = np.zeros((self.n_parameters, self.n_dims), dtype=bool)
    full_mask[slice(s + self.n_dims, s + 2 * self.n_dims)] = dim_mask
    jac[:, full_mask] = 1
    return jac


def similarity_d_dp(sim, points):
    r"""
    Computes the Jacobian of the transform w.r.t the parameters.

    The Jacobian generated (for 2D) is of the form::

        x -y 1 0
        y  x 0 1

    This maintains a parameter order of::

      W(x;p) = [1 + a  -b   ] [x] + tx
               [b      1 + a] [y] + ty

    Parameters
    ----------
    points : (n_points, n_dims) ndarray
        The set of points to calculate the jacobian for.

    Returns
    -------
    (n_points, n_params, n_dims) ndarray
        The jacobian wrt parametrization

    Raises
    ------
    DimensionalityError
        `points.n_dims != self.n_dims` or transform is not 2D

    """
    n_points, points_n_dim = points.shape
    if points_n_dim != sim.n_dims:
        raise ValueError('Trying to sample jacobian in incorrect '
                         'dimensions (transform is {0}D, sampling '
                         'at {1}D)'.format(sim.n_dims, points_n_dim))
    elif sim.n_dims != 2:
        # TODO: implement 3D Jacobian
        raise ValueError("Only the Jacobian of a 2D similarity "
                         "transform is currently supported.")

    # prealloc the jacobian
    jac = np.zeros((n_points, sim.n_parameters, sim.n_dims))
    ones = np.ones_like(points)

    # Build a mask and apply it to the points to build the jacobian
    # Do this for each parameter - [a, b, tx, ty] respectively
    _apply_jacobian_mask(sim, jac, np.array([1, 1]), 0, points)
    _apply_jacobian_mask(sim, jac, np.array([-1, 1]), 1, points[:, ::-1])
    _apply_jacobian_mask(sim, jac, np.array([1, 0]), 2, ones)
    _apply_jacobian_mask(sim, jac, np.array([0, 1]), 3, ones)

    return jac


def _apply_jacobian_mask(sim, jac, param_mask, row_index, points):
    # make a mask for a single points jacobian
    full_mask = np.zeros((sim.n_parameters, sim.n_dims), dtype=np.bool)
    # fill the mask in for the ith axis
    full_mask[row_index] = [True, True]
    # assign the ith axis points to this mask, broadcasting over all
    # points
    jac[:, full_mask] = points * param_mask
